@@ -4,7 +4,9 @@
 #include "arch/arch.h"
 #include "ast/async_event_types.h"
 #include "bpftrace.h"
+#include "codegen_helper.h"
 #include "irbuilderbpf.h"
+#include "log.h"
 #include "utils.h"
 
 #include <llvm/IR/DataLayout.h>
@@ -18,6 +20,68 @@ namespace libbpf {
 namespace bpftrace {
 namespace ast {
 
+namespace {
+std::string probeReadHelperName(libbpf::bpf_func_id id)
+{
+  switch (id)
+  {
+    case libbpf::BPF_FUNC_probe_read:
+      return "probe_read";
+    case libbpf::BPF_FUNC_probe_read_user:
+      return "probe_read_user";
+    case libbpf::BPF_FUNC_probe_read_kernel:
+      return "probe_read_kernel";
+    case libbpf::BPF_FUNC_probe_read_str:
+      return "probe_read_str";
+    case libbpf::BPF_FUNC_probe_read_user_str:
+      return "probe_read_user_str";
+    case libbpf::BPF_FUNC_probe_read_kernel_str:
+      return "probe_read_kernel_str";
+    default:
+      throw std::runtime_error("BUG: unknown probe_read id: " +
+                               std::to_string(id));
+  }
+}
+} // namespace
+
+libbpf::bpf_func_id IRBuilderBPF::selectProbeReadHelper(AddrSpace as, bool str)
+{
+  libbpf::bpf_func_id fn;
+  // Assume that if a kernel has probe_read_kernel it has the other 3 too
+  if (bpftrace_.feature_->has_helper_probe_read_kernel())
+  {
+    if (as == AddrSpace::kernel)
+    {
+      fn = str ? libbpf::BPF_FUNC_probe_read_kernel_str
+               : libbpf::BPF_FUNC_probe_read_kernel;
+    }
+    else if (as == AddrSpace::user)
+    {
+      fn = str ? libbpf::BPF_FUNC_probe_read_user_str
+               : libbpf::BPF_FUNC_probe_read_user;
+    }
+    else
+    {
+      // if the kernel has the new helpers but AS is still none it is a bug
+      // in bpftrace, assert catches it for debug builds.
+      // assert(as != AddrSpace::none);
+      static bool warnonce = false;
+      if (!warnonce)
+      {
+        warnonce = true;
+        LOG(WARNING) << "Addrspace is not set";
+      }
+      fn = str ? libbpf::BPF_FUNC_probe_read_str : libbpf::BPF_FUNC_probe_read;
+    }
+  }
+  else
+  {
+    fn = str ? libbpf::BPF_FUNC_probe_read_str : libbpf::BPF_FUNC_probe_read;
+  }
+
+  return fn;
+}
+
 AllocaInst *IRBuilderBPF::CreateUSym(llvm::Value *val)
 {
   std::vector<llvm::Type *> elements = {
@@ -71,7 +135,7 @@ IRBuilderBPF::IRBuilderBPF(LLVMContext &context,
       &module_);
 }
 
-AllocaInst *IRBuilderBPF::CreateAllocaBPF(llvm::Type *ty, llvm::Value *arraysize, const std::string &name)
+void IRBuilderBPF::hoist(const std::function<void()> &functor)
 {
   Function *parent = GetInsertBlock()->getParent();
   BasicBlock &entry_block = parent->getEntryBlock();
@@ -81,8 +145,19 @@ AllocaInst *IRBuilderBPF::CreateAllocaBPF(llvm::Type *ty, llvm::Value *arraysize
     SetInsertPoint(&entry_block);
   else
     SetInsertPoint(&entry_block.front());
-  AllocaInst *alloca = CreateAlloca(ty, arraysize, name);
+
+  functor();
   restoreIP(ip);
+}
+
+AllocaInst *IRBuilderBPF::CreateAllocaBPF(llvm::Type *ty,
+                                          llvm::Value *arraysize,
+                                          const std::string &name)
+{
+  AllocaInst *alloca;
+  hoist([this, ty, arraysize, &name, &alloca]() {
+    alloca = CreateAlloca(ty, arraysize, name);
+  });
 
   CreateLifetimeStart(alloca);
   return alloca;
@@ -101,30 +176,20 @@ AllocaInst *IRBuilderBPF::CreateAllocaBPF(const SizedType &stype, const std::str
 
 AllocaInst *IRBuilderBPF::CreateAllocaBPFInit(const SizedType &stype, const std::string &name)
 {
-  Function *parent = GetInsertBlock()->getParent();
-  BasicBlock &entry_block = parent->getEntryBlock();
-
-  auto ip = saveIP();
-  if (entry_block.empty())
-    SetInsertPoint(&entry_block);
-  else
-    SetInsertPoint(&entry_block.front());
-
-  llvm::Type *ty = GetType(stype);
-  AllocaInst *alloca = CreateAllocaBPF(ty, nullptr, name);
-
-  if (!stype.IsAggregate())
-  {
-    CreateStore(getInt64(0), alloca);
-  }
-  else
-  {
-    CREATE_MEMSET(alloca, getInt8(0), stype.size, 1);
-  }
-
-  restoreIP(ip);
-
-  CreateLifetimeStart(alloca);
+  AllocaInst *alloca;
+  hoist([this, &stype, &name, &alloca]() {
+    llvm::Type *ty = GetType(stype);
+    alloca = CreateAlloca(ty, nullptr, name);
+    CreateLifetimeStart(alloca);
+    if (needMemcpy(stype))
+    {
+      CREATE_MEMSET(alloca, getInt8(0), stype.GetSize(), 1);
+    }
+    else
+    {
+      CreateStore(ConstantInt::get(ty, 0), alloca);
+    }
+  });
   return alloca;
 }
 
@@ -156,27 +221,36 @@ llvm::ConstantInt *IRBuilderBPF::GetIntSameSize(uint64_t C, llvm::Value *expr)
 llvm::Type *IRBuilderBPF::GetType(const SizedType &stype)
 {
   llvm::Type *ty;
-  if (stype.IsArray())
+  if (stype.IsByteArray() || stype.IsRecordTy())
   {
-    ty = ArrayType::get(getInt8Ty(), stype.size);
+    ty = ArrayType::get(getInt8Ty(), stype.GetSize());
+  }
+  else if (stype.IsArrayTy())
+  {
+    ty = ArrayType::get(GetType(*stype.GetElementTy()), stype.GetNumElements());
   }
   else if (stype.IsTupleTy())
   {
     std::vector<llvm::Type *> llvm_elems;
     std::ostringstream ty_name;
 
-    for (const auto &elemtype : stype.tuple_elems)
+    for (const auto &elem : stype.GetFields())
     {
+      auto &elemtype = elem.type;
       llvm_elems.emplace_back(GetType(elemtype));
       ty_name << elemtype << "_";
     }
     ty_name << "_tuple_t";
 
-    ty = GetStructType(ty_name.str(), llvm_elems, true);
+    ty = GetStructType(ty_name.str(), llvm_elems, false);
+  }
+  else if (stype.IsPtrTy())
+  {
+    ty = getInt64Ty();
   }
   else
   {
-    switch (stype.size)
+    switch (stype.GetSize())
     {
       case 16:
         ty = getInt128Ty();
@@ -194,8 +268,8 @@ llvm::Type *IRBuilderBPF::GetType(const SizedType &stype)
         ty = getInt8Ty();
         break;
       default:
-        std::cerr << stype.size << " is not a valid type size for GetType" << std::endl;
-        abort();
+        LOG(FATAL) << stype.GetSize()
+                   << " is not a valid type size for GetType";
     }
   }
   return ty;
@@ -214,23 +288,37 @@ CallInst *IRBuilderBPF::createCall(Value *callee,
 #endif
 }
 
-CallInst *IRBuilderBPF::CreateBpfPseudoCall(int mapfd)
+CallInst *IRBuilderBPF::CreateBpfPseudoCallId(int mapid)
 {
   Function *pseudo_func = module_.getFunction("llvm.bpf.pseudo");
   return createCall(pseudo_func,
-                    { getInt64(BPF_PSEUDO_MAP_FD), getInt64(mapfd) },
+                    { getInt64(BPF_PSEUDO_MAP_FD), getInt64(mapid) },
                     "pseudo");
 }
 
-CallInst *IRBuilderBPF::CreateBpfPseudoCall(Map &map)
+CallInst *IRBuilderBPF::CreateBpfPseudoCallId(Map &map)
 {
-  int mapfd = bpftrace_.maps_[map.ident]->mapfd_;
-  return CreateBpfPseudoCall(mapfd);
+  int mapid = bpftrace_.maps[map.ident].value()->id;
+  return CreateBpfPseudoCallId(mapid);
 }
 
-CallInst *IRBuilderBPF::createMapLookup(int mapfd, AllocaInst *key)
+CallInst *IRBuilderBPF::CreateBpfPseudoCallValue(int mapid)
 {
-  Value *map_ptr = CreateBpfPseudoCall(mapfd);
+  Function *pseudo_func = module_.getFunction("llvm.bpf.pseudo");
+  return CreateCall(pseudo_func,
+                    { getInt64(BPF_PSEUDO_MAP_VALUE), getInt64(mapid) },
+                    "pseudo");
+}
+
+CallInst *IRBuilderBPF::CreateBpfPseudoCallValue(Map &map)
+{
+  int mapid = bpftrace_.maps[map.ident].value()->id;
+  return CreateBpfPseudoCallValue(mapid);
+}
+
+CallInst *IRBuilderBPF::createMapLookup(int mapid, Value *key)
+{
+  Value *map_ptr = CreateBpfPseudoCallId(mapid);
   // void *map_lookup_elem(struct bpf_map * map, void * key)
   // Return: Map value or NULL
 
@@ -250,29 +338,30 @@ CallInst *IRBuilderBPF::CreateGetJoinMap(Value *ctx, const location &loc)
   AllocaInst *key = CreateAllocaBPF(getInt32Ty(), "key");
   CreateStore(getInt32(0), key);
 
-  CallInst *call = createMapLookup(bpftrace_.join_map_->mapfd_, key);
+  CallInst *call = createMapLookup(
+      bpftrace_.maps[MapManager::Type::Join].value()->id, key);
   CreateHelperErrorCond(ctx, call, libbpf::BPF_FUNC_map_lookup_elem, loc, true);
   return call;
 }
 
 Value *IRBuilderBPF::CreateMapLookupElem(Value *ctx,
                                          Map &map,
-                                         AllocaInst *key,
+                                         Value *key,
                                          const location &loc)
 {
   assert(ctx && ctx->getType() == getInt8PtrTy());
-  int mapfd = bpftrace_.maps_[map.ident]->mapfd_;
-  return CreateMapLookupElem(ctx, mapfd, key, map.type, loc);
+  int mapid = bpftrace_.maps[map.ident].value()->id;
+  return CreateMapLookupElem(ctx, mapid, key, map.type, loc);
 }
 
 Value *IRBuilderBPF::CreateMapLookupElem(Value *ctx,
-                                         int mapfd,
-                                         AllocaInst *key,
+                                         int mapid,
+                                         Value *key,
                                          SizedType &type,
                                          const location &loc)
 {
   assert(ctx && ctx->getType() == getInt8PtrTy());
-  CallInst *call = createMapLookup(mapfd, key);
+  CallInst *call = createMapLookup(mapid, key);
 
   // Check if result == 0
   Function *parent = GetInsertBlock()->getParent();
@@ -288,8 +377,8 @@ Value *IRBuilderBPF::CreateMapLookupElem(Value *ctx,
   CreateCondBr(condition, lookup_success_block, lookup_failure_block);
 
   SetInsertPoint(lookup_success_block);
-  if (type.IsAggregate())
-    CREATE_MEMCPY(value, call, type.size, 1);
+  if (needMemcpy(type))
+    CREATE_MEMCPY(value, call, type.GetSize(), 1);
   else
   {
     assert(value->getType()->isPointerTy() &&
@@ -301,27 +390,29 @@ Value *IRBuilderBPF::CreateMapLookupElem(Value *ctx,
   CreateBr(lookup_merge_block);
 
   SetInsertPoint(lookup_failure_block);
-  if (type.IsAggregate())
-    CREATE_MEMSET(value, getInt8(0), type.size, 1);
+  if (needMemcpy(type))
+    CREATE_MEMSET(value, getInt8(0), type.GetSize(), 1);
   else
     CreateStore(getInt64(0), value);
   CreateHelperError(ctx, getInt32(0), libbpf::BPF_FUNC_map_lookup_elem, loc);
   CreateBr(lookup_merge_block);
 
   SetInsertPoint(lookup_merge_block);
-  if (type.IsAggregate())
+  if (needMemcpy(type))
     return value;
 
-  return CreateLoad(value);
+  Value *ret = CreateLoad(value);
+  CreateLifetimeEnd(value);
+  return ret;
 }
 
 void IRBuilderBPF::CreateMapUpdateElem(Value *ctx,
                                        Map &map,
-                                       AllocaInst *key,
+                                       Value *key,
                                        Value *val,
                                        const location &loc)
 {
-  Value *map_ptr = CreateBpfPseudoCall(map);
+  Value *map_ptr = CreateBpfPseudoCallId(map);
 
   assert(ctx && ctx->getType() == getInt8PtrTy());
   assert(key->getType()->isPointerTy());
@@ -348,12 +439,12 @@ void IRBuilderBPF::CreateMapUpdateElem(Value *ctx,
 
 void IRBuilderBPF::CreateMapDeleteElem(Value *ctx,
                                        Map &map,
-                                       AllocaInst *key,
+                                       Value *key,
                                        const location &loc)
 {
   assert(ctx && ctx->getType() == getInt8PtrTy());
   assert(key->getType()->isPointerTy());
-  Value *map_ptr = CreateBpfPseudoCall(map);
+  Value *map_ptr = CreateBpfPseudoCallId(map);
 
   // int map_delete_elem(&map, &key)
   // Return: 0 on success or negative error
@@ -369,35 +460,46 @@ void IRBuilderBPF::CreateMapDeleteElem(Value *ctx,
 }
 
 void IRBuilderBPF::CreateProbeRead(Value *ctx,
-                                   AllocaInst *dst,
+                                   Value *dst,
                                    size_t size,
                                    Value *src,
+                                   AddrSpace as,
                                    const location &loc)
 {
-  return CreateProbeRead(ctx, dst, getInt32(size), src, loc);
+  return CreateProbeRead(ctx, dst, getInt32(size), src, as, loc);
 }
 
 void IRBuilderBPF::CreateProbeRead(Value *ctx,
-                                   AllocaInst *dst,
+                                   Value *dst,
                                    llvm::Value *size,
                                    Value *src,
+                                   AddrSpace as,
                                    const location &loc)
 {
   assert(ctx && ctx->getType() == getInt8PtrTy());
+  assert(size && size->getType()->getIntegerBitWidth() <= 32);
+  size = CreateIntCast(size, getInt32Ty(), false);
+
   // int bpf_probe_read(void *dst, int size, void *src)
   // Return: 0 on success or negative error
+
+  auto read_fn = selectProbeReadHelper(as, false);
+
   FunctionType *proberead_func_type = FunctionType::get(
       getInt64Ty(), { dst->getType(), getInt32Ty(), src->getType() }, false);
   PointerType *proberead_func_ptr_type = PointerType::get(proberead_func_type, 0);
-  Constant *proberead_func = ConstantExpr::getCast(
-      Instruction::IntToPtr,
-      getInt64(libbpf::BPF_FUNC_probe_read),
-      proberead_func_ptr_type);
-  CallInst *call = createCall(proberead_func, { dst, size, src }, "probe_read");
-  CreateHelperErrorCond(ctx, call, libbpf::BPF_FUNC_probe_read, loc);
+  Constant *proberead_func = ConstantExpr::getCast(Instruction::IntToPtr,
+                                                   getInt64(read_fn),
+                                                   proberead_func_ptr_type);
+  CallInst *call = createCall(proberead_func,
+                              { dst, size, src },
+                              probeReadHelperName(read_fn));
+  CreateHelperErrorCond(ctx, call, read_fn, loc);
 }
 
-Constant *IRBuilderBPF::createProbeReadStrFn(llvm::Type *dst, llvm::Type *src)
+Constant *IRBuilderBPF::createProbeReadStrFn(llvm::Type *dst,
+                                             llvm::Type *src,
+                                             AddrSpace as)
 {
   assert(src && (src->isIntegerTy() || src->isPointerTy()));
   // int bpf_probe_read_str(void *dst, int size, const void *unsafe_ptr)
@@ -406,7 +508,7 @@ Constant *IRBuilderBPF::createProbeReadStrFn(llvm::Type *dst, llvm::Type *src)
   PointerType *probereadstr_func_ptr_type = PointerType::get(
       probereadstr_func_type, 0);
   return ConstantExpr::getCast(Instruction::IntToPtr,
-                               getInt64(libbpf::BPF_FUNC_probe_read_str),
+                               getInt64(selectProbeReadHelper(as, true)),
                                probereadstr_func_ptr_type);
 }
 
@@ -414,24 +516,27 @@ CallInst *IRBuilderBPF::CreateProbeReadStr(Value *ctx,
                                            AllocaInst *dst,
                                            size_t size,
                                            Value *src,
+                                           AddrSpace as,
                                            const location &loc)
 {
   assert(ctx && ctx->getType() == getInt8PtrTy());
-  return CreateProbeReadStr(ctx, dst, getInt32(size), src, loc);
+  return CreateProbeReadStr(ctx, dst, getInt32(size), src, as, loc);
 }
 
 CallInst *IRBuilderBPF::CreateProbeReadStr(Value *ctx,
                                            Value *dst,
                                            size_t size,
                                            Value *src,
+                                           AddrSpace as,
                                            const location &loc)
 {
   assert(ctx && ctx->getType() == getInt8PtrTy());
-  Constant *fn = createProbeReadStrFn(dst->getType(), src->getType());
+  Constant *fn = createProbeReadStrFn(dst->getType(), src->getType(), as);
+  auto read_fn = selectProbeReadHelper(as, true);
   CallInst *call = createCall(fn,
                               { dst, getInt32(size), src },
-                              "probe_read_str");
-  CreateHelperErrorCond(ctx, call, libbpf::BPF_FUNC_probe_read_str, loc);
+                              probeReadHelperName(read_fn));
+  CreateHelperErrorCond(ctx, call, read_fn, loc);
   return call;
 }
 
@@ -439,6 +544,7 @@ CallInst *IRBuilderBPF::CreateProbeReadStr(Value *ctx,
                                            AllocaInst *dst,
                                            llvm::Value *size,
                                            Value *src,
+                                           AddrSpace as,
                                            const location &loc)
 {
   assert(ctx && ctx->getType() == getInt8PtrTy());
@@ -448,66 +554,109 @@ CallInst *IRBuilderBPF::CreateProbeReadStr(Value *ctx,
 
   auto *size_i32 = CreateIntCast(size, getInt32Ty(), false);
 
-  Constant *fn = createProbeReadStrFn(dst->getType(), src->getType());
-  CallInst *call = createCall(fn, { dst, size_i32, src }, "probe_read_str");
-  CreateHelperErrorCond(ctx, call, libbpf::BPF_FUNC_probe_read_str, loc);
+  Constant *fn = createProbeReadStrFn(dst->getType(), src->getType(), as);
+  auto read_fn = selectProbeReadHelper(as, true);
+  CallInst *call = createCall(fn,
+                              { dst, size_i32, src },
+                              probeReadHelperName(read_fn));
+  CreateHelperErrorCond(ctx, call, read_fn, loc);
   return call;
 }
 
 Value *IRBuilderBPF::CreateUSDTReadArgument(Value *ctx,
                                             struct bcc_usdt_argument *argument,
                                             Builtin &builtin,
+                                            AddrSpace as,
                                             const location &loc)
 {
   assert(ctx && ctx->getType() == getInt8PtrTy());
-  // TODO (mmarchini): Handle base + index * scale addressing.
-  // https://github.com/iovisor/bcc/pull/988
-  if (argument->valid & BCC_USDT_ARGUMENT_INDEX_REGISTER_NAME)
-    std::cerr << "index register is not handled yet [" << argument->index_register_name << "]" << std::endl;
-  if (argument->valid & BCC_USDT_ARGUMENT_SCALE)
-    std::cerr << "scale is not handled yet [" << argument->scale << "]" << std::endl;
+  // Argument size must be 1, 2, 4, or 8. See
+  // https://sourceware.org/systemtap/wiki/UserSpaceProbeImplementation
+  int abs_size = std::abs(argument->size);
+  assert(abs_size == 1 || abs_size == 2 || abs_size == 4 || abs_size == 8);
   if (argument->valid & BCC_USDT_ARGUMENT_DEREF_IDENT)
-    std::cerr << "defer ident is not handled yet [" << argument->deref_ident << "]" << std::endl;
+    LOG(ERROR) << "deref ident is not handled yet [" << argument->deref_ident
+               << "]";
+  // USDT arguments can be any valid gas (GNU asm) operand.
+  // BCC normalises these into the bcc_usdt_argument and supports most
+  // valid gas operands.
+  //
+  // This code handles the following argument types:
+  // * A constant (ARGUMENT_CONSTANT)
+  //
+  // * The value of a register (ARGUMENT_BASE_REGISTER_NAME without
+  // ARGUMENT_DEREF_OFFSET set).
+  //
+  // * The value at address: base_register + offset + (index_register * scale)
+  // Where index_register and scale are optional.
+  // Note: Offset is optional in the gas operand, however will be set as zero
+  // if the register needs to be dereferenced.
 
   if (argument->valid & BCC_USDT_ARGUMENT_CONSTANT)
-    return getInt64(argument->constant);
+  {
+    // Correctly sign extend and convert to a 64-bit int
+    return CreateIntCast(getIntN(abs_size * 8, argument->constant),
+                         getInt64Ty(),
+                         argument->size < 0);
+  }
 
+  if (argument->valid & BCC_USDT_ARGUMENT_INDEX_REGISTER_NAME &&
+      !(argument->valid & BCC_USDT_ARGUMENT_BASE_REGISTER_NAME))
+  {
+    // Invalid combination??
+    LOG(ERROR) << "index register set without base register;"
+               << " this case is not yet handled";
+  }
   Value *result = nullptr;
   if (argument->valid & BCC_USDT_ARGUMENT_BASE_REGISTER_NAME) {
     int offset = 0;
     offset = arch::offset(argument->base_register_name);
     if (offset < 0)
     {
-      std::cerr << "offset for register " << argument->base_register_name
-                << " not known" << std::endl;
-      abort();
+      LOG(FATAL) << "offset for register " << argument->base_register_name
+                 << " not known";
     }
 
-    // Argument size must be 1, 2, 4, or 8. See
-    // https://sourceware.org/systemtap/wiki/UserSpaceProbeImplementation
-    int abs_size = std::abs(argument->size);
-    assert(abs_size == 1 || abs_size == 2 || abs_size == 4 || abs_size == 8);
 
     // bpftrace's args are internally represented as 64 bit integers. However,
     // the underlying argument (of the target program) may be less than 64
     // bits. So we must be careful to zero out unused bits.
     Value* reg = CreateGEP(ctx, getInt64(offset * sizeof(uintptr_t)), "load_register");
     AllocaInst *dst = CreateAllocaBPF(builtin.type, builtin.ident);
+    Value *index_offset = nullptr;
+    if (argument->valid & BCC_USDT_ARGUMENT_INDEX_REGISTER_NAME)
+    {
+      int ioffset = arch::offset(argument->index_register_name);
+      if (ioffset < 0)
+      {
+        LOG(FATAL) << "offset for register " << argument->index_register_name
+                   << " not known";
+      }
+      index_offset = CreateGEP(ctx,
+                               getInt64(ioffset * sizeof(uintptr_t)),
+                               "load_register");
+      index_offset = CreateLoad(getInt64Ty(), index_offset);
+      if (argument->valid & BCC_USDT_ARGUMENT_SCALE)
+      {
+        index_offset = CreateMul(index_offset, getInt64(argument->scale));
+      }
+    }
     if (argument->valid & BCC_USDT_ARGUMENT_DEREF_OFFSET) {
       Value *ptr = CreateAdd(CreateLoad(getInt64Ty(), reg),
                              getInt64(argument->deref_offset));
-      // Zero out `dst` here in case we read less than 64 bits
-      CreateStore(getInt64(0), dst);
-      CreateProbeRead(ctx, dst, abs_size, ptr, loc);
-      result = CreateLoad(dst);
+      if (index_offset)
+      {
+        ptr = CreateAdd(ptr, index_offset);
+      }
+      CreateProbeRead(ctx, dst, abs_size, ptr, as, loc);
+      result = CreateLoad(getIntNTy(abs_size * 8), dst);
     }
     else
     {
-      // Zero out `dst` in case we read less than 64 bits
-      CreateStore(getInt64(0), dst);
-      CreateProbeRead(ctx, dst, abs_size, reg, loc);
-      result = CreateLoad(dst);
+      result = CreateLoad(getIntNTy(abs_size * 8), reg);
     }
+    // Sign extend and convert to a bpftools standard 64-bit integer type
+    result = CreateIntCast(result, getInt64Ty(), argument->size < 0);
     CreateLifetimeEnd(dst);
   }
   return result;
@@ -519,6 +668,7 @@ Value *IRBuilderBPF::CreateUSDTReadArgument(Value *ctx,
                                             int arg_num,
                                             Builtin &builtin,
                                             pid_t pid,
+                                            AddrSpace as,
                                             const location &loc)
 {
   assert(ctx && ctx->getType() == getInt8PtrTy());
@@ -534,7 +684,8 @@ Value *IRBuilderBPF::CreateUSDTReadArgument(Value *ctx,
   }
 
   if (usdt == nullptr) {
-    std::cerr << "failed to initialize usdt context for probe " << attach_point->target << std::endl;
+    LOG(ERROR) << "failed to initialize usdt context for probe "
+               << attach_point->target;
     exit(-1);
   }
 
@@ -548,86 +699,21 @@ Value *IRBuilderBPF::CreateUSDTReadArgument(Value *ctx,
                             arg_num,
                             &argument) != 0)
   {
-    std::cerr << "couldn't get argument " << arg_num << " for " << attach_point->target << ":"
-              << attach_point->ns << ":" << attach_point->func << std::endl;
+    LOG(ERROR) << "couldn't get argument " << arg_num << " for "
+               << attach_point->target << ":" << attach_point->ns << ":"
+               << attach_point->func;
     exit(-2);
   }
 
-  Value *result = CreateUSDTReadArgument(ctx, &argument, builtin, loc);
+  Value *result = CreateUSDTReadArgument(ctx, &argument, builtin, as, loc);
 
   bcc_usdt_close(usdt);
   return result;
 }
 
-Value *IRBuilderBPF::CreateStrcmp(Value *ctx,
-                                  Value *val,
-                                  std::string str,
-                                  const location &loc,
-                                  bool inverse)
-{
-  assert(ctx && ctx->getType() == getInt8PtrTy());
-  auto cmpAmount = strlen(str.c_str()) + 1;
-  return CreateStrncmp(ctx, val, str, cmpAmount, loc, inverse);
-}
-
-Value *IRBuilderBPF::CreateStrncmp(Value *ctx __attribute__((unused)),
-                                   Value *val,
-                                   std::string str,
-                                   uint64_t n,
-                                   const location &loc __attribute__((unused)),
-                                   bool inverse)
-{
-  assert(ctx && ctx->getType() == getInt8PtrTy());
-#ifndef NDEBUG
-  PointerType *valp = cast<PointerType>(val->getType());
-  assert(valp->getElementType()->isArrayTy() &&
-         valp->getElementType()->getArrayNumElements() >= n &&
-         valp->getElementType()->getArrayElementType() == getInt8Ty());
-#endif
-
-  Function *parent = GetInsertBlock()->getParent();
-  BasicBlock *str_ne = BasicBlock::Create(module_.getContext(), "strcmp.false", parent);
-  AllocaInst *store = CreateAllocaBPF(getInt8Ty(), "strcmp.result");
-
-  CreateStore(getInt1(!inverse), store);
-
-  const char *c_str = str.c_str();
-  for (size_t i = 0; i < n; i++)
-  {
-    BasicBlock *char_eq = BasicBlock::Create(module_.getContext(), "strcmp.loop", parent);
-
-    auto *ptr = CreateGEP(val, { getInt32(0), getInt32(i) });
-    Value *l = CreateLoad(getInt8Ty(), ptr);
-    Value *r = getInt8(c_str[i]);
-    Value *cmp = CreateICmpNE(l, r, "strcmp.cmp");
-    CreateCondBr(cmp, str_ne, char_eq);
-    SetInsertPoint(char_eq);
-  }
-  CreateStore(getInt1(inverse), store);
-  CreateBr(str_ne);
-
-  SetInsertPoint(str_ne);
-  Value *result = CreateLoad(store);
-  CreateLifetimeEnd(store);
-  result = CreateIntCast(result, getInt64Ty(), false);
-  return result;
-}
-
-Value *IRBuilderBPF::CreateStrcmp(Value *ctx,
-                                  Value *val1,
-                                  Value *val2,
-                                  const location &loc,
-                                  bool inverse)
-{
-  assert(ctx && ctx->getType() == getInt8PtrTy());
-  return CreateStrncmp(ctx, val1, val2, bpftrace_.strlen_, loc, inverse);
-}
-
-Value *IRBuilderBPF::CreateStrncmp(Value *ctx,
-                                   Value *val1,
+Value *IRBuilderBPF::CreateStrncmp(Value *val1,
                                    Value *val2,
                                    uint64_t n,
-                                   const location &loc,
                                    bool inverse)
 {
   /*
@@ -652,19 +738,42 @@ Value *IRBuilderBPF::CreateStrncmp(Value *ctx,
      }
   */
 
+  // Check if the compared strings are literals.
+  // If so, we can avoid storing the literal in memory.
+  std::optional<std::string> literal1;
+  if (auto constString1 = dyn_cast<ConstantDataArray>(val1))
+    literal1 = constString1->getAsString();
+  else if (isa<ConstantAggregateZero>(val1))
+    literal1 = "";
+  else
+    literal1 = std::nullopt;
+
+  std::optional<std::string> literal2;
+  if (auto constString2 = dyn_cast<ConstantDataArray>(val2))
+    literal2 = constString2->getAsString();
+  else if (isa<ConstantAggregateZero>(val2))
+    literal2 = "";
+  else
+    literal2 = std::nullopt;
+
 #ifndef NDEBUG
-  PointerType *val1p = cast<PointerType>(val1->getType());
-  PointerType *val2p = cast<PointerType>(val2->getType());
+  if (!literal1)
+  {
+    PointerType *val1p = cast<PointerType>(val1->getType());
+    assert(val1p->getElementType()->isArrayTy() &&
+           val1p->getElementType()->getArrayElementType() == getInt8Ty());
+  }
+  if (!literal2)
+  {
+    PointerType *val2p = cast<PointerType>(val2->getType());
 
-  assert(val1p->getElementType()->isArrayTy() &&
-         val1p->getElementType()->getArrayElementType() == getInt8Ty());
-
-  assert(val2p->getElementType()->isArrayTy() &&
-         val2p->getElementType()->getArrayElementType() == getInt8Ty());
+    assert(val2p->getElementType()->isArrayTy() &&
+           val2p->getElementType()->getArrayElementType() == getInt8Ty());
+  }
 #endif
 
   Function *parent = GetInsertBlock()->getParent();
-  AllocaInst *store = CreateAllocaBPF(getInt8Ty(), "strcmp.result");
+  AllocaInst *store = CreateAllocaBPF(getInt1Ty(), "strcmp.result");
   BasicBlock *str_ne = BasicBlock::Create(module_.getContext(),
                                           "strcmp.false",
                                           parent);
@@ -675,9 +784,6 @@ Value *IRBuilderBPF::CreateStrncmp(Value *ctx,
   CreateStore(getInt1(!inverse), store);
 
   Value *null_byte = getInt8(0);
-
-  AllocaInst *val_l = CreateAllocaBPF(getInt8Ty(), "strcmp.char_l");
-  AllocaInst *val_r = CreateAllocaBPF(getInt8Ty(), "strcmp.char_r");
   for (size_t i = 0; i < n; i++)
   {
     BasicBlock *char_eq = BasicBlock::Create(module_.getContext(),
@@ -687,13 +793,23 @@ Value *IRBuilderBPF::CreateStrncmp(Value *ctx,
                                                      "strcmp.loop_null_cmp",
                                                      parent);
 
-    auto *ptr1 = CreateGEP(val1, { getInt32(0), getInt32(i) });
-    CreateProbeRead(ctx, val_l, 1, ptr1, loc);
-    Value *l = CreateLoad(getInt8Ty(), val_l);
+    Value *l;
+    if (literal1)
+      l = getInt8(literal1->c_str()[i]);
+    else
+    {
+      auto *ptr_l = CreateGEP(val1, { getInt32(0), getInt32(i) });
+      l = CreateLoad(getInt8Ty(), ptr_l);
+    }
 
-    auto *ptr2 = CreateGEP(val2, { getInt32(0), getInt32(i) });
-    CreateProbeRead(ctx, val_r, 1, ptr2, loc);
-    Value *r = CreateLoad(getInt8Ty(), val_r);
+    Value *r;
+    if (literal2)
+      r = getInt8(literal2->c_str()[i]);
+    else
+    {
+      auto *ptr_r = CreateGEP(val2, { getInt32(0), getInt32(i) });
+      r = CreateLoad(getInt8Ty(), ptr_r);
+    }
 
     Value *cmp = CreateICmpNE(l, r, "strcmp.cmp");
     CreateCondBr(cmp, str_ne, loop_null_check);
@@ -715,23 +831,24 @@ Value *IRBuilderBPF::CreateStrncmp(Value *ctx,
 
   Value *result = CreateLoad(store);
   CreateLifetimeEnd(store);
-  CreateLifetimeEnd(val_l);
-  CreateLifetimeEnd(val_r);
   result = CreateIntCast(result, getInt64Ty(), false);
 
   return result;
 }
 
-CallInst *IRBuilderBPF::CreateGetNs()
+CallInst *IRBuilderBPF::CreateGetNs(bool boot_time)
 {
   // u64 ktime_get_ns()
   // Return: current ktime
+
+  auto fn = boot_time ? libbpf::BPF_FUNC_ktime_get_boot_ns
+                      : libbpf::BPF_FUNC_ktime_get_ns;
+
   FunctionType *gettime_func_type = FunctionType::get(getInt64Ty(), false);
   PointerType *gettime_func_ptr_type = PointerType::get(gettime_func_type, 0);
-  Constant *gettime_func = ConstantExpr::getCast(
-      Instruction::IntToPtr,
-      getInt64(libbpf::BPF_FUNC_ktime_get_ns),
-      gettime_func_ptr_type);
+  Constant *gettime_func = ConstantExpr::getCast(Instruction::IntToPtr,
+                                                 getInt64(fn),
+                                                 gettime_func_ptr_type);
   return createCall(gettime_func, {}, "get_ns");
 }
 
@@ -820,10 +937,9 @@ CallInst *IRBuilderBPF::CreateGetStackId(Value *ctx,
                                          const location &loc)
 {
   assert(ctx && ctx->getType() == getInt8PtrTy());
-  assert(bpftrace_.stackid_maps_.count(stack_type) == 1);
 
-  Value *map_ptr = CreateBpfPseudoCall(
-      bpftrace_.stackid_maps_[stack_type]->mapfd_);
+  Value *map_ptr = CreateBpfPseudoCallId(
+      bpftrace_.maps[stack_type].value()->id);
 
   int flags = 0;
   if (ustack)
@@ -878,9 +994,10 @@ void IRBuilderBPF::CreatePerfEventOutput(Value *ctx, Value *data, size_t size)
   assert(ctx && ctx->getType() == getInt8PtrTy());
   assert(data && data->getType()->isPointerTy());
 
-  Value *map_ptr = CreateBpfPseudoCall(bpftrace_.perf_event_map_->mapfd_);
+  Value *map_ptr = CreateBpfPseudoCallId(
+      bpftrace_.maps[MapManager::Type::PerfEvent].value()->id);
 
-  Value *flags_val = CreateGetCpuId();
+  Value *flags_val = getInt64(BPF_F_CURRENT_CPU);
   Value *size_val = getInt64(size);
 
   // int bpf_perf_event_output(struct pt_regs *ctx, struct bpf_map *map,
@@ -937,9 +1054,10 @@ Value *IRBuilderBPF::CreatKFuncArg(Value *ctx,
                                    SizedType &type,
                                    std::string &name)
 {
+  assert(type.IsIntTy() || type.IsPtrTy());
   ctx = CreatePointerCast(ctx, getInt64Ty()->getPointerTo());
-  Value *expr = CreateLoad(getInt64Ty(),
-                           CreateGEP(ctx, getInt64(type.kfarg_idx)),
+  Value *expr = CreateLoad(GetType(type),
+                           CreateGEP(ctx, getInt64(type.funcarg_idx)),
                            name);
 
   // LLVM 7.0 <= does not have CreateLoad(*Ty, *Ptr, isVolatile, Name),
@@ -948,6 +1066,29 @@ Value *IRBuilderBPF::CreatKFuncArg(Value *ctx,
   return expr;
 }
 
+Value *IRBuilderBPF::CreateRegisterRead(Value *ctx, const std::string &builtin)
+{
+  int offset;
+  if (builtin == "retval")
+    offset = arch::ret_offset();
+  else if (builtin == "func")
+    offset = arch::pc_offset();
+  else // argX
+    offset = arch::arg_offset(atoi(builtin.substr(3).c_str()));
+
+  Value *ctx_ptr = CreatePointerCast(ctx, getInt64Ty()->getPointerTo());
+  // LLVM optimization is possible to transform `(uint64*)ctx` into
+  // `(uint8*)ctx`, but sometimes this causes invalid context access.
+  // Mark every context access to suppress any LLVM optimization.
+  Value *result = CreateLoad(getInt64Ty(),
+                             CreateGEP(ctx_ptr, getInt64(offset)),
+                             builtin);
+  // LLVM 7.0 <= does not have CreateLoad(*Ty, *Ptr, isVolatile, Name),
+  // so call setVolatile() manually
+  dyn_cast<LoadInst>(result)->setVolatile(true);
+  return result;
+}
+
 static bool return_zero_if_err(libbpf::bpf_func_id func_id)
 {
   switch (func_id)
@@ -985,7 +1126,8 @@ void IRBuilderBPF::CreateHelperError(Value *ctx,
     return;
 
   int error_id = helper_error_id_++;
-  bpftrace_.helper_error_info_[error_id] = { .func_id = func_id, .loc = loc };
+  bpftrace_.resources.helper_error_info[error_id] = { .func_id = func_id,
+                                                      .loc = loc };
 
   auto elements = AsyncEvent::HelperError().asLLVMType(*this);
   StructType *helper_error_struct = GetStructType("helper_error_t",
@@ -1041,5 +1183,74 @@ void IRBuilderBPF::CreateHelperErrorCond(Value *ctx,
   SetInsertPoint(helper_merge_block);
 }
 
+void IRBuilderBPF::CreatePath(Value *ctx,
+                              AllocaInst *buf,
+                              Value *path,
+                              const location &loc)
+{
+  // int bpf_d_path(struct path *path, char *buf, u32 sz)
+  // Return: 0 or error
+  FunctionType *d_path_func_type = FunctionType::get(
+      getInt64Ty(), { getInt8PtrTy(), buf->getType(), getInt32Ty() }, false);
+  PointerType *d_path_func_ptr_type = PointerType::get(d_path_func_type, 0);
+  Constant *d_path_func = ConstantExpr::getCast(Instruction::IntToPtr,
+                                                getInt64(
+                                                    libbpf::BPF_FUNC_d_path),
+                                                d_path_func_ptr_type);
+  CallInst *call = createCall(d_path_func,
+                              { path, buf, getInt32(bpftrace_.strlen_) },
+                              "d_path");
+  CreateHelperErrorCond(ctx, call, libbpf::BPF_FUNC_d_path, loc);
+}
+
+void IRBuilderBPF::CreateSeqPrintf(Value *ctx,
+                                   Value *fmt,
+                                   Value *fmt_size,
+                                   AllocaInst *data,
+                                   Value *data_len,
+                                   const location &loc)
+{
+  // long bpf_seq_printf(struct seq_file *m, const char *fmt, __u32 fmt_size,
+  //                     const void *data, __u32 data_len)
+  // Return: 0 or error
+  FunctionType *seq_printf_func_type = FunctionType::get(getInt64Ty(),
+                                                         { getInt64Ty(),
+                                                           getInt8PtrTy(),
+                                                           getInt32Ty(),
+                                                           getInt8PtrTy(),
+                                                           getInt32Ty() },
+                                                         false);
+  PointerType *seq_printf_func_ptr_type = PointerType::get(seq_printf_func_type,
+                                                           0);
+  Constant *seq_printf_func = ConstantExpr::getCast(
+      Instruction::IntToPtr,
+      getInt64(libbpf::BPF_FUNC_seq_printf),
+      seq_printf_func_ptr_type);
+
+  ctx = CreatePointerCast(ctx, getInt8Ty()->getPointerTo());
+  Value *meta = CreateLoad(getInt64Ty()->getPointerTo(),
+                           CreateGEP(ctx, getInt64(0)),
+                           "meta");
+  dyn_cast<LoadInst>(meta)->setVolatile(true);
+
+  Value *seq = CreateLoad(getInt64Ty(), CreateGEP(meta, getInt64(0)), "seq");
+
+  CallInst *call = createCall(seq_printf_func,
+                              { seq, fmt, fmt_size, data, data_len },
+                              "seq_printf");
+  CreateHelperErrorCond(ctx, call, libbpf::BPF_FUNC_seq_printf, loc);
+}
+
+StoreInst *IRBuilderBPF::createAlignedStore(Value *val,
+                                            Value *ptr,
+                                            unsigned int align)
+{
+#if LLVM_VERSION_MAJOR < 10
+  return CreateAlignedStore(val, ptr, align);
+#else
+  return CreateAlignedStore(val, ptr, MaybeAlign(align));
+#endif
+}
+
 } // namespace ast
 } // namespace bpftrace
