@@ -1,28 +1,36 @@
+#ifndef _GNU_SOURCE
+#define _GNU_SOURCE
+#endif
+
 #include <cstring>
 #include <elf.h>
 #include <fcntl.h>
 #include <fstream>
 #include <iostream>
-#include <link.h>
 #include <linux/hw_breakpoint.h>
 #include <linux/limits.h>
 #include <linux/perf_event.h>
 #include <regex>
-#include <sys/auxv.h>
+#include <sys/ioctl.h>
+#include <sys/syscall.h>
 #include <sys/utsname.h>
 #include <tuple>
 #include <unistd.h>
 
-#include "attached_probe.h"
-#include "bpftrace.h"
-#include "disasm.h"
-#include "list.h"
-#include "usdt.h"
 #include <bcc/bcc_elf.h>
 #include <bcc/bcc_syms.h>
 #include <bcc/bcc_usdt.h>
+
+#include "attached_probe.h"
+#include "bpftrace.h"
+#include "disasm.h"
+#include "log.h"
+#include "probe_matcher.h"
+#include "usdt.h"
+#ifdef HAVE_LIBBPF_BPF_H
+#include <bpf/bpf.h>
+#endif
 #include <linux/perf_event.h>
-#include <linux/version.h>
 
 namespace libbpf {
 #undef __BPF_FUNC_MAPPER
@@ -51,9 +59,9 @@ bpf_probe_attach_type attachtype(ProbeType t)
     case ProbeType::uretprobe: return BPF_PROBE_RETURN; break;
     case ProbeType::usdt:      return BPF_PROBE_ENTRY;  break;
     default:
-      std::cerr << "invalid probe attachtype \"" << probetypeName(t) << "\"" << std::endl;
-      abort();
+      LOG(FATAL) << "invalid probe attachtype \"" << probetypeName(t) << "\"";
   }
+  // lgtm[cpp/missing-return]
 }
 
 bpf_prog_type progtype(ProbeType t)
@@ -66,10 +74,17 @@ bpf_prog_type progtype(ProbeType t)
     case ProbeType::uretprobe:  return BPF_PROG_TYPE_KPROBE; break;
     case ProbeType::usdt:       return BPF_PROG_TYPE_KPROBE; break;
     case ProbeType::tracepoint: return BPF_PROG_TYPE_TRACEPOINT; break;
-    case ProbeType::profile:      return BPF_PROG_TYPE_PERF_EVENT; break;
-    case ProbeType::interval:      return BPF_PROG_TYPE_PERF_EVENT; break;
+    case ProbeType::profile:
+      return BPF_PROG_TYPE_PERF_EVENT;
+      break;
+    case ProbeType::interval:
+      return BPF_PROG_TYPE_PERF_EVENT;
+      break;
     case ProbeType::software:   return BPF_PROG_TYPE_PERF_EVENT; break;
     case ProbeType::watchpoint: return BPF_PROG_TYPE_PERF_EVENT; break;
+    case ProbeType::asyncwatchpoint:
+      return BPF_PROG_TYPE_PERF_EVENT;
+      break;
     case ProbeType::hardware:   return BPF_PROG_TYPE_PERF_EVENT; break;
     case ProbeType::kfunc:
       return static_cast<enum ::bpf_prog_type>(libbpf::BPF_PROG_TYPE_TRACING);
@@ -77,30 +92,36 @@ bpf_prog_type progtype(ProbeType t)
     case ProbeType::kretfunc:
       return static_cast<enum ::bpf_prog_type>(libbpf::BPF_PROG_TYPE_TRACING);
       break;
-    default:
-      std::cerr << "program type not found" << std::endl;
-      abort();
+    case ProbeType::iter:
+      return static_cast<enum ::bpf_prog_type>(libbpf::BPF_PROG_TYPE_TRACING);
+      break;
+    case ProbeType::invalid:
+      LOG(FATAL) << "program type invalid";
   }
+
+  return {}; // unreached
 }
 
 std::string progtypeName(bpf_prog_type t)
 {
-  switch (t)
+  switch (static_cast<libbpf::bpf_prog_type>(t))
   {
     // clang-format off
-    case BPF_PROG_TYPE_KPROBE:     return "BPF_PROG_TYPE_KPROBE";     break;
-    case BPF_PROG_TYPE_TRACEPOINT: return "BPF_PROG_TYPE_TRACEPOINT"; break;
-    case BPF_PROG_TYPE_PERF_EVENT: return "BPF_PROG_TYPE_PERF_EVENT"; break;
+    case libbpf::BPF_PROG_TYPE_KPROBE:     return "BPF_PROG_TYPE_KPROBE";     break;
+    case libbpf::BPF_PROG_TYPE_TRACEPOINT: return "BPF_PROG_TYPE_TRACEPOINT"; break;
+    case libbpf::BPF_PROG_TYPE_PERF_EVENT: return "BPF_PROG_TYPE_PERF_EVENT"; break;
+    case libbpf::BPF_PROG_TYPE_TRACING:    return "BPF_PROG_TYPE_TRACING";    break;
     // clang-format on
     default:
-      std::cerr << "invalid program type: " << t << std::endl;
-      abort();
+      LOG(FATAL) << "invalid program type: " << t;
   }
+  // lgtm[cpp/missing-return]
 }
 
 void check_banned_kretprobes(std::string const& kprobe_name) {
   if (banned_kretprobes.find(kprobe_name) != banned_kretprobes.end()) {
-    std::cerr << "error: kretprobe:" << kprobe_name << " can't be used as it might lock up your system." << std::endl;
+    LOG(ERROR) << "error: kretprobe:" << kprobe_name
+               << " can't be used as it might lock up your system.";
     exit(1);
   }
 }
@@ -128,12 +149,45 @@ void AttachedProbe::attach_kfunc(void)
 
 int AttachedProbe::detach_kfunc(void)
 {
-  std::cerr << "kfunc not available for linked against bcc version"
-            << std::endl;
+  LOG(ERROR) << "kfunc not available for linked against bcc version";
   return -1;
 }
 #endif // HAVE_BCC_KFUNC
 
+#ifdef HAVE_LIBBPF_LINK_CREATE
+void AttachedProbe::attach_iter(void)
+{
+  linkfd_ = bpf_link_create(progfd_,
+                            0,
+                            static_cast<enum ::bpf_attach_type>(
+                                libbpf::BPF_TRACE_ITER),
+                            NULL);
+  if (linkfd_ < 0)
+  {
+    throw std::runtime_error("Error attaching probe: '" + probe_.name + "'");
+  }
+}
+
+int AttachedProbe::detach_iter(void)
+{
+  close(linkfd_);
+  return 0;
+}
+#else
+void AttachedProbe::attach_iter(void)
+{
+  throw std::runtime_error(
+      "Error attaching probe: " + probe_.name +
+      ", iter API is not available for linked libbpf version");
+}
+
+int AttachedProbe::detach_iter(void)
+{
+  LOG(ERROR) << "iter is not available for linked bpf version";
+  return 0;
+}
+#endif // HAVE_LIBBPF_LINK_CREATE
+
 AttachedProbe::AttachedProbe(Probe &probe, std::tuple<uint8_t *, uintptr_t> func, bool safe_mode)
   : probe_(probe), func_(func)
 {
@@ -172,27 +226,34 @@ AttachedProbe::AttachedProbe(Probe &probe, std::tuple<uint8_t *, uintptr_t> func
     case ProbeType::kretfunc:
       attach_kfunc();
       break;
+    case ProbeType::iter:
+      attach_iter();
+      break;
     default:
-      std::cerr << "invalid attached probe type \"" << probetypeName(probe_.type) << "\"" << std::endl;
-      abort();
+      LOG(FATAL) << "invalid attached probe type \""
+                 << probetypeName(probe_.type) << "\"";
   }
 }
 
-AttachedProbe::AttachedProbe(Probe &probe, std::tuple<uint8_t *, uintptr_t> func, int pid)
-  : probe_(probe), func_(func)
+AttachedProbe::AttachedProbe(Probe &probe,
+                             std::tuple<uint8_t *, uintptr_t> func,
+                             int pid,
+                             BPFfeature &feature)
+    : probe_(probe), func_(func)
 {
   load_prog();
   switch (probe_.type)
   {
     case ProbeType::usdt:
-      attach_usdt(pid);
+      attach_usdt(pid, feature);
       break;
     case ProbeType::watchpoint:
+    case ProbeType::asyncwatchpoint:
       attach_watchpoint(pid, probe.mode);
       break;
     default:
-      std::cerr << "invalid attached probe type \"" << probetypeName(probe_.type) << "\"" << std::endl;
-      abort();
+      LOG(FATAL) << "invalid attached probe type \""
+                 << probetypeName(probe_.type) << "\"";
   }
 }
 
@@ -203,7 +264,7 @@ AttachedProbe::~AttachedProbe()
   {
     err = bpf_close_perf_event_fd(perf_event_fd);
     if (err)
-      std::cerr << "Error closing perf event FDs for probe: " << probe_.name << std::endl;
+      LOG(ERROR) << "failed to close perf event FDs for probe: " << probe_.name;
   }
 
   err = 0;
@@ -217,6 +278,9 @@ AttachedProbe::~AttachedProbe()
     case ProbeType::kretfunc:
       err = detach_kfunc();
       break;
+    case ProbeType::iter:
+      err = detach_iter();
+      break;
     case ProbeType::uprobe:
     case ProbeType::uretprobe:
     case ProbeType::usdt:
@@ -231,19 +295,25 @@ AttachedProbe::~AttachedProbe()
     case ProbeType::interval:
     case ProbeType::software:
     case ProbeType::watchpoint:
+    case ProbeType::asyncwatchpoint:
     case ProbeType::hardware:
       break;
-    default:
-      std::cerr << "invalid attached probe type \"" << probetypeName(probe_.type) << "\" at destructor" << std::endl;
-      abort();
+    case ProbeType::invalid:
+      LOG(FATAL) << "invalid attached probe type \""
+                 << probetypeName(probe_.type) << "\" at destructor";
   }
   if (err)
-    std::cerr << "Error detaching probe: " << probe_.name << std::endl;
+    LOG(ERROR) << "failed to detach probe: " << probe_.name;
 
   if (progfd_ >= 0)
     close(progfd_);
 }
 
+const Probe &AttachedProbe::probe() const
+{
+  return probe_;
+}
+
 std::string AttachedProbe::eventprefix() const
 {
   switch (attachtype(probe_.type))
@@ -252,10 +322,9 @@ std::string AttachedProbe::eventprefix() const
       return "p_";
     case BPF_PROBE_RETURN:
       return "r_";
-    default:
-      std::cerr << "invalid eventprefix" << std::endl;
-      abort();
   }
+
+  return {}; // unreached
 }
 
 std::string AttachedProbe::eventname() const
@@ -277,8 +346,8 @@ std::string AttachedProbe::eventname() const
     case ProbeType::tracepoint:
       return probe_.attach_point;
     default:
-      std::cerr << "invalid eventname probe \"" << probetypeName(probe_.type) << "\"" << std::endl;
-      abort();
+      LOG(FATAL) << "invalid eventname probe \"" << probetypeName(probe_.type)
+                 << "\"";
   }
 }
 
@@ -330,6 +399,11 @@ resolve_offset(const std::string &path, const std::string &symbol, uint64_t loc)
   if (bcc_resolve_symname(path.c_str(), symbol.c_str(), loc, 0, nullptr, &bcc_sym))
      throw std::runtime_error("Could not resolve symbol: " + path + ":" + symbol);
 
+  // Have to free sym.module, see:
+  // https://github.com/iovisor/bcc/blob/ba73657cb8c4dab83dfb89eed4a8b3866255569a/src/cc/bcc_syms.h#L98-L99
+  if (bcc_sym.module)
+    ::free(const_cast<char *>(bcc_sym.module));
+
   return bcc_sym.offset;
 }
 
@@ -346,25 +420,24 @@ static void check_alignment(std::string &path,
 
   std::string tmp = path + ":" + symbol + "+" + std::to_string(func_offset);
 
-  if (AlignState::Ok == aligned)
-    return;
-
-    // If we did not allow unaligned uprobes in the
-    // compile time, force the safe mode now.
+  // If we did not allow unaligned uprobes in the
+  // compile time, force the safe mode now.
 #ifndef HAVE_UNSAFE_PROBE
   safe_mode = true;
 #endif
 
   switch (aligned)
   {
+    case AlignState::Ok:
+      return;
     case AlignState::NotAlign:
       if (safe_mode)
         throw std::runtime_error("Could not add " + probe_name +
                                  " into middle of instruction: " + tmp);
       else
-        std::cerr << "Unsafe " + probe_name +
-                         " in the middle of the instruction: "
-                  << tmp << std::endl;
+        LOG(WARNING) << "Unsafe " + probe_name +
+                            " in the middle of the instruction: "
+                     << tmp;
       break;
 
     case AlignState::Fail:
@@ -372,7 +445,7 @@ static void check_alignment(std::string &path,
         throw std::runtime_error("Failed to check if " + probe_name +
                                  " is in proper place: " + tmp);
       else
-        std::cerr << "Unchecked " + probe_name + ": " << tmp << std::endl;
+        LOG(WARNING) << "Unchecked " + probe_name + ": " << tmp;
       break;
 
     case AlignState::NotSupp:
@@ -382,11 +455,8 @@ static void check_alignment(std::string &path,
                                  "(k|u)probe offset support): " +
                                  tmp);
       else
-        std::cerr << "Unchecked " + probe_name + " : " << tmp << std::endl;
+        LOG(WARNING) << "Unchecked " + probe_name + " : " << tmp;
       break;
-
-    default:
-      throw std::runtime_error("Internal error: " + tmp);
   }
 }
 
@@ -417,11 +487,10 @@ void AttachedProbe::resolve_offset_uprobe(bool safe_mode)
       }
       else
       {
-        std::cerr << "WARNING: could not determine instruction boundary for "
-                  << probe_.name
-                  << " (binary appears stripped). Misaligned probes "
-                     "can lead to tracee crashes!"
-                  << std::endl;
+        LOG(WARNING) << "Could not determine instruction boundary for "
+                     << probe_.name
+                     << " (binary appears stripped). Misaligned probes "
+                        "can lead to tracee crashes!";
         offset_ = probe_.address;
         return;
       }
@@ -512,7 +581,7 @@ void AttachedProbe::resolve_offset_kprobe(bool safe_mode)
   sym.name = symbol;
   const struct vmlinux_location *locs = vmlinux_locs;
   struct vmlinux_location locs_env[] = {
-    { nullptr, true },
+    { nullptr, false },
     { nullptr, false },
   };
   char *env_path = std::getenv("BPFTRACE_VMLINUX");
@@ -558,96 +627,6 @@ void AttachedProbe::resolve_offset_kprobe(bool safe_mode)
       path, symbol, sym_offset, func_offset, safe_mode, probe_.type);
 }
 
-/**
- * Search for LINUX_VERSION_CODE in the vDSO, returning 0 if it can't be found.
- */
-static unsigned _find_version_note(unsigned long base)
-{
-  auto ehdr = reinterpret_cast<const ElfW(Ehdr) *>(base);
-
-  for (int i = 0; i < ehdr->e_shnum; i++)
-  {
-    auto shdr = reinterpret_cast<const ElfW(Shdr) *>(
-      base + ehdr->e_shoff + (i * ehdr->e_shentsize)
-    );
-
-    if (shdr->sh_type == SHT_NOTE)
-    {
-      auto ptr = reinterpret_cast<const char *>(base + shdr->sh_offset);
-      auto end = ptr + shdr->sh_size;
-
-      while (ptr < end)
-      {
-        auto nhdr = reinterpret_cast<const ElfW(Nhdr) *>(ptr);
-        ptr += sizeof *nhdr;
-
-        auto name = ptr;
-        ptr += (nhdr->n_namesz + sizeof(ElfW(Word)) - 1) & -sizeof(ElfW(Word));
-
-        auto desc = ptr;
-        ptr += (nhdr->n_descsz + sizeof(ElfW(Word)) - 1) & -sizeof(ElfW(Word));
-
-        if ((nhdr->n_namesz > 5 && !memcmp(name, "Linux", 5)) &&
-            nhdr->n_descsz == 4 && !nhdr->n_type)
-          return *reinterpret_cast<const uint32_t *>(desc);
-      }
-    }
-  }
-
-  return 0;
-}
-
-/**
- * Find a LINUX_VERSION_CODE matching the host kernel. The build-time constant
- * may not match if bpftrace is compiled on a different Linux version than it's
- * used on, e.g. if built with Docker.
- */
-static unsigned kernel_version(int attempt)
-{
-  switch (attempt)
-  {
-    case 0:
-    {
-      // Fetch LINUX_VERSION_CODE from the vDSO .note section, falling back on
-      // the build-time constant if unavailable. This always matches the
-      // running kernel, but is not supported on arm32.
-      unsigned code = 0;
-      unsigned long base = getauxval(AT_SYSINFO_EHDR);
-      if (base && !memcmp(reinterpret_cast<void *>(base), ELFMAG, 4))
-        code = _find_version_note(base);
-      if (! code)
-        code = LINUX_VERSION_CODE;
-      return code;
-    }
-    case 1:
-      struct utsname utsname;
-      if (uname(&utsname) < 0)
-        return 0;
-      unsigned x, y, z;
-      if (sscanf(utsname.release, "%u.%u.%u", &x, &y, &z) != 3)
-        return 0;
-      return KERNEL_VERSION(x, y, z);
-    case 2:
-    {
-      // Try to get the definition of LINUX_VERSION_CODE at runtime.
-      std::ifstream linux_version_header{"/usr/include/linux/version.h"};
-      const std::string content{std::istreambuf_iterator<char>(linux_version_header),
-                                std::istreambuf_iterator<char>()};
-      const std::regex regex{"#define\\s+LINUX_VERSION_CODE\\s+(\\d+)"};
-      std::smatch match;
-
-      if (std::regex_search(content.begin(), content.end(), match, regex))
-        return static_cast<unsigned>(std::stoi(match[1]));
-
-      return 0;
-    }
-    default:
-      break;
-  }
-  std::cerr << "invalid kernel version" << std::endl;
-  abort();
-}
-
 void AttachedProbe::load_prog()
 {
   uint8_t *insns = std::get<0>(func_);
@@ -686,6 +665,9 @@ void AttachedProbe::load_prog()
     tracing_type = probetypeName(probe_.type);
     if (!tracing_type.empty())
     {
+      if (tracing_type == "iter")
+        tracing_type = "bpf_iter";
+
       tracing_name = tracing_type + "__" + namep;
       namep = tracing_name.c_str();
     }
@@ -702,6 +684,27 @@ void AttachedProbe::load_prog()
         continue;
       }
 
+#ifdef HAVE_BCC_PROG_LOAD_XATTR
+      struct bpf_load_program_attr attr = {};
+
+      attr.prog_type = progtype(probe_.type);
+      attr.name = namep;
+      attr.insns = reinterpret_cast<struct bpf_insn *>(insns);
+      attr.license = license;
+
+      libbpf::bpf_prog_type prog_type = static_cast<libbpf::bpf_prog_type>(
+          progtype(probe_.type));
+
+      if (prog_type != libbpf::BPF_PROG_TYPE_TRACING &&
+          prog_type != libbpf::BPF_PROG_TYPE_EXT)
+        attr.kern_version = version;
+
+      attr.log_level = log_level;
+
+      progfd_ = bcc_prog_load_xattr(
+          &attr, prog_len, log_buf.get(), log_buf_size, true);
+
+#else // HAVE_BCC_PROG_LOAD_XATTR
 #ifdef HAVE_BCC_PROG_LOAD
       progfd_ = bcc_prog_load(progtype(probe_.type),
                               namep,
@@ -716,6 +719,8 @@ void AttachedProbe::load_prog()
                               log_level,
                               log_buf.get(),
                               log_buf_size);
+#endif // HAVE_BCC_PROG_LOAD_XATTR
+
       if (progfd_ >= 0)
         break;
     }
@@ -749,7 +754,7 @@ void AttachedProbe::load_prog()
       std::cout << std::endl << "Program ID: " << info.id << std::endl;
     }
     std::cout << std::endl
-              << "Bytecode: " << std::endl
+              << "The verifier log: " << std::endl
               << log_buf.get() << std::endl;
   }
 }
@@ -776,7 +781,7 @@ void AttachedProbe::attach_kprobe(bool safe_mode)
     if (probe_.orig_name != probe_.name) {
       // a wildcard expansion couldn't probe something, just print a warning
       // as this is normal for some kernel functions (eg, do_debug())
-      std::cerr << "Warning: could not attach probe " << probe_.name << ", skipping." << std::endl;
+      LOG(WARNING) << "could not attach probe " << probe_.name << ", skipping.";
     } else {
       // an explicit match failed, so fail as the user must have wanted it
       throw std::runtime_error("Error attaching probe: '" + probe_.name + "'");
@@ -814,16 +819,36 @@ void AttachedProbe::attach_uprobe(bool safe_mode)
   perf_event_fds_.push_back(perf_event_fd);
 }
 
-void AttachedProbe::attach_usdt(int pid)
+int AttachedProbe::usdt_sem_up_manual(const std::string &fn_name, void *ctx)
 {
-  struct bcc_usdt_location loc = {};
   int err;
-  void *ctx;
-  // TODO: fn_name may need a unique suffix for each attachment on the same
-  // probe:
-  std::string fn_name = "probe_" + probe_.attach_point + "_1";
+
+#ifdef BCC_USDT_HAS_FULLY_SPECIFIED_PROBE
+  if (probe_.ns == "")
+    err = bcc_usdt_enable_probe(ctx,
+                                probe_.attach_point.c_str(),
+                                fn_name.c_str());
+  else
+    err = bcc_usdt_enable_fully_specified_probe(
+        ctx, probe_.ns.c_str(), probe_.attach_point.c_str(), fn_name.c_str());
+#else
+  err = bcc_usdt_enable_probe(ctx,
+                              probe_.attach_point.c_str(),
+                              fn_name.c_str());
+#endif // BCC_USDT_HAS_FULLY_SPECIFIED_PROBE
+
+  // Defer context destruction until probes are detached b/c context
+  // destruction will decrement usdt semaphore count.
+  usdt_destructor_ = [ctx]() { bcc_usdt_close(ctx); };
+
+  return err;
+}
 
 #ifdef HAVE_BCC_USDT_ADDSEM
+int AttachedProbe::usdt_sem_up_manual_addsem(int pid,
+                                             const std::string &fn_name,
+                                             void *ctx)
+{
   // NB: we are careful to capture by value here everything that will not
   // be available in AttachedProbe destructor.
   auto addsem = [this, fn_name](void *c, int16_t val) -> int {
@@ -838,55 +863,116 @@ void AttachedProbe::attach_usdt(int pid)
           fn_name.c_str(),
           val);
   };
-#endif // HAVE_BCC_USDT_ADDSEM
 
-  if (pid)
-  {
-    // FIXME when iovisor/bcc#2064 is merged, optionally pass probe_.path
-    ctx = bcc_usdt_new_frompid(pid, nullptr);
-    if (!ctx)
-      throw std::runtime_error("Error initializing context for probe: " + probe_.name + ", for PID: " + std::to_string(pid));
+  // Set destructor to decrement the semaphore count
+  usdt_destructor_ = [pid, addsem]() {
+    void *c = bcc_usdt_new_frompid(pid, nullptr);
+    if (!c)
+      return;
 
-#ifdef HAVE_BCC_USDT_ADDSEM
-    usdt_destructor_ = [pid, addsem]() {
-      void *c = bcc_usdt_new_frompid(pid, nullptr);
-      if (!c)
-        return;
+    addsem(c, -1);
+    bcc_usdt_close(c);
+  };
 
-      addsem(c, -1);
-      bcc_usdt_close(c);
-    };
-#endif // HAVE_BCC_USDT_ADDSEM
-  }
-  else
-  {
-    ctx = bcc_usdt_new_frompath(probe_.path.c_str());
-    if (!ctx)
-      throw std::runtime_error("Error initializing context for probe: " + probe_.name);
-  }
-
-#ifndef HAVE_BCC_USDT_ADDSEM
-  // Defer context destruction until probes are detached b/c context
-  // destruction will decrement usdt semaphore count.
-  usdt_destructor_ = [ctx]() { bcc_usdt_close(ctx); };
-#endif // HAVE_BCC_USDT_ADDSEM
-
-#ifdef HAVE_BCC_USDT_ADDSEM
   // Use semaphore increment API to avoid having to hold onto the usdt context
   // for the entire tracing session. Reason we do it this way instead of
   // holding onto usdt context is b/c each usdt context can take lots of memory
   // (~10MB). This, coupled with --usdt-file-activation and tracees that have a
   // forking model can cause bpftrace to use huge amounts of memory if we hold
   // onto the contexts.
-  err = addsem(ctx, +1);
-#elif defined(BCC_USDT_HAS_FULLY_SPECIFIED_PROBE)
-  if (probe_.ns == "")
-    err = bcc_usdt_enable_probe(ctx, probe_.attach_point.c_str(), fn_name.c_str());
-  else
-    err = bcc_usdt_enable_fully_specified_probe(ctx, probe_.ns.c_str(), probe_.attach_point.c_str(), fn_name.c_str());
+  int err = addsem(ctx, +1);
+
+  // Now close the context to save some memory
+  bcc_usdt_close(ctx);
+
+  return err;
+}
 #else
-  err = bcc_usdt_enable_probe(ctx, probe_.attach_point.c_str(), fn_name.c_str());
+int AttachedProbe::usdt_sem_up_manual_addsem(int pid __attribute__((unused)),
+                                             const std::string &fn_name
+                                             __attribute__((unused)),
+                                             void *ctx __attribute__((unused)))
+{
+  return 0;
+}
+#endif // HAVE_BCC_USDT_ADDSEM
+
+int AttachedProbe::usdt_sem_up([[maybe_unused]] BPFfeature &feature,
+                               [[maybe_unused]] int pid,
+                               const std::string &fn_name,
+                               void *ctx)
+{
+  // If we have BCC and kernel support for uprobe refcnt API, then we don't
+  // need to do anything here. The kernel will increment the semaphore count
+  // for us when we provide the semaphore offset.
+  if (feature.has_uprobe_refcnt())
+  {
+    bcc_usdt_close(ctx);
+    return 0;
+  }
+
+#if defined(HAVE_BCC_USDT_ADDSEM)
+  return usdt_sem_up_manual_addsem(pid, fn_name, ctx);
+#else
+  return usdt_sem_up_manual(fn_name, ctx);
 #endif
+}
+
+void AttachedProbe::attach_usdt(int pid, BPFfeature &feature)
+{
+  struct bcc_usdt_location loc = {};
+  int err;
+  void *ctx;
+  // TODO: fn_name may need a unique suffix for each attachment on the same
+  // probe:
+  std::string fn_name = "probe_" + probe_.attach_point + "_1";
+
+  if (pid)
+  {
+    // FIXME when iovisor/bcc#2064 is merged, optionally pass probe_.path
+    ctx = bcc_usdt_new_frompid(pid, nullptr);
+    if (!ctx)
+      throw std::runtime_error(
+          "Error initializing context for probe: " + probe_.name +
+          ", for PID: " + std::to_string(pid));
+  }
+  else
+  {
+    ctx = bcc_usdt_new_frompath(probe_.path.c_str());
+    if (!ctx)
+      throw std::runtime_error("Error initializing context for probe: " +
+                               probe_.name);
+  }
+
+  // Resolve location of usdt probe
+  auto u = USDTHelper::find(pid, probe_.path, probe_.ns, probe_.attach_point);
+  if (!u.has_value())
+    throw std::runtime_error("Failed to find usdt probe: " + eventname());
+  probe_.path = u->path;
+
+  err = bcc_usdt_get_location(ctx,
+                              probe_.ns.c_str(),
+                              probe_.attach_point.c_str(),
+                              probe_.usdt_location_idx,
+                              &loc);
+  if (err)
+    throw std::runtime_error("Error finding location for probe: " +
+                             probe_.name);
+  probe_.loc = loc.address;
+
+  offset_ = resolve_offset(probe_.path, probe_.attach_point, probe_.loc);
+
+  // Should be 0 if there's no semaphore
+  //
+  // Cast to 32 bits b/c kernel API only takes 32 bit offset
+  [[maybe_unused]] auto semaphore_offset = static_cast<uint32_t>(
+      u->semaphore_offset);
+
+  // Increment the semaphore count (will noop if no semaphore)
+  //
+  // NB: Do *not* use `ctx` after this call. It may either be open or closed,
+  // depending on which path was taken.
+  err = usdt_sem_up(feature, pid, fn_name, ctx);
 
   if (err)
   {
@@ -897,24 +983,6 @@ void AttachedProbe::attach_usdt(int pid)
     throw std::runtime_error(err);
   }
 
-  auto u = USDTHelper::find(pid, probe_.path, probe_.ns, probe_.attach_point);
-  if (!u.has_value())
-    throw std::runtime_error("Failed to find usdt probe: " + eventname());
-  probe_.path = u->path;
-
-  err = bcc_usdt_get_location(ctx, probe_.ns.c_str(), probe_.attach_point.c_str(), 0, &loc);
-  if (err)
-    throw std::runtime_error("Error finding location for probe: " + probe_.name);
-  probe_.loc = loc.address;
-
-#ifdef HAVE_BCC_USDT_ADDSEM
-  // If we use the bcc_usdt_addsem*() API, bcc won't decrement semaphore count
-  // in bcc_usdt_close(). So we are free to close context here.
-  bcc_usdt_close(ctx);
-#endif // HAVE_BCC_USDT_ADDSEM
-
-  offset_ = resolve_offset(probe_.path, probe_.attach_point, probe_.loc);
-
   int perf_event_fd =
 #ifdef LIBBCC_ATTACH_UPROBE_SEVEN_ARGS_SIGNATURE
       bpf_attach_uprobe(progfd_,
@@ -923,7 +991,7 @@ void AttachedProbe::attach_usdt(int pid)
                         probe_.path.c_str(),
                         offset_,
                         pid == 0 ? -1 : pid,
-                        0);
+                        semaphore_offset);
 #else
       bpf_attach_uprobe(progfd_,
                         attachtype(probe_.type),
@@ -936,7 +1004,8 @@ void AttachedProbe::attach_usdt(int pid)
   if (perf_event_fd < 0)
   {
     if (pid)
-      throw std::runtime_error("Error attaching probe: " + probe_.name + ", to PID: " + std::to_string(pid));
+      throw std::runtime_error("Error attaching probe: " + probe_.name +
+                               ", to PID: " + std::to_string(pid));
     else
       throw std::runtime_error("Error attaching probe: " + probe_.name);
   }
@@ -983,8 +1052,7 @@ void AttachedProbe::attach_profile()
   }
   else
   {
-    std::cerr << "invalid profile path \"" << probe_.path << "\"" << std::endl;
-    abort();
+    LOG(FATAL) << "invalid profile path \"" << probe_.path << "\"";
   }
 
   std::vector<int> cpus = get_online_cpus();
@@ -1025,8 +1093,7 @@ void AttachedProbe::attach_interval()
   }
   else
   {
-    std::cerr << "invalid interval path \"" << probe_.path << "\"" << std::endl;
-    abort();
+    LOG(FATAL) << "invalid interval path \"" << probe_.path << "\"";
   }
 
   int perf_event_fd = bpf_attach_perf_event(progfd_,
@@ -1116,10 +1183,6 @@ void AttachedProbe::attach_hardware()
 
 void AttachedProbe::attach_watchpoint(int pid, const std::string& mode)
 {
-  if (pid < 1) {
-    throw std::runtime_error("pid not provided for " + probe_.name);
-  }
-
   struct perf_event_attr attr = {};
   attr.type = PERF_TYPE_BREAKPOINT;
   attr.size = sizeof(struct perf_event_attr);
@@ -1140,11 +1203,50 @@ void AttachedProbe::attach_watchpoint(int pid, const std::string& mode)
   // Generate a notification every 1 event; we care about every event
   attr.sample_period = 1;
 
-  int perf_event_fd = bpf_attach_perf_event_raw(progfd_, &attr, pid, -1, -1, 0);
-  if (perf_event_fd < 0)
-    throw std::runtime_error("Error attaching probe: " + probe_.name);
+  std::vector<int> cpus;
+  if (pid >= 1)
+  {
+    cpus = { -1 };
+  }
+  else
+  {
+    cpus = get_online_cpus();
+    pid = -1;
+  }
 
-  perf_event_fds_.push_back(perf_event_fd);
+  for (int cpu : cpus)
+  {
+    // We copy paste the code from bcc's bpf_attach_perf_event_raw here
+    // because we need to know the exact error codes (and also we don't
+    // want bcc's noisy error messages).
+    int perf_event_fd = syscall(
+        __NR_perf_event_open, &attr, pid, cpu, -1, PERF_FLAG_FD_CLOEXEC);
+    if (perf_event_fd < 0)
+    {
+      if (errno == ENOSPC)
+        throw EnospcException("No more HW registers left");
+      else
+        throw std::system_error(errno,
+                                std::generic_category(),
+                                "Error attaching probe: " + probe_.name);
+    }
+    if (ioctl(perf_event_fd, PERF_EVENT_IOC_SET_BPF, progfd_) != 0)
+    {
+      close(perf_event_fd);
+      throw std::system_error(errno,
+                              std::generic_category(),
+                              "Error attaching probe: " + probe_.name);
+    }
+    if (ioctl(perf_event_fd, PERF_EVENT_IOC_ENABLE, 0) != 0)
+    {
+      close(perf_event_fd);
+      throw std::system_error(errno,
+                              std::generic_category(),
+                              "Error attaching probe: " + probe_.name);
+    }
+
+    perf_event_fds_.push_back(perf_event_fd);
+  }
 }
 
 } // namespace bpftrace
