@@ -4,6 +4,7 @@
 #include <linux/types.h>
 #include <map>
 #include <regex>
+#include <set>
 #include <string>
 #include <unistd.h>
 #include <unordered_set>
@@ -13,6 +14,8 @@ struct btf_type;
 
 namespace bpftrace {
 
+class BPFtrace;
+
 class BTF
 {
   enum state
@@ -23,14 +26,21 @@ class BTF
 
 public:
   BTF();
+  BTF(BPFtrace* bpftrace) : BTF()
+  {
+    bpftrace_ = bpftrace;
+  };
   ~BTF();
 
   bool has_data(void) const;
   std::string c_def(const std::unordered_set<std::string>& set) const;
   std::string type_of(const std::string& name, const std::string& field);
   std::string type_of(const btf_type* type, const std::string& field);
-  void display_funcs(std::regex* re) const;
-  void display_structs(std::regex* re) const;
+
+  std::set<std::string> get_all_structs() const;
+  std::unique_ptr<std::istream> get_all_funcs() const;
+  std::map<std::string, std::vector<std::string>> get_params(
+      const std::set<std::string>& funcs) const;
 
   int resolve_args(const std::string &func,
                    std::map<std::string, SizedType>& args,
@@ -42,6 +52,7 @@ private:
 
   struct btf* btf;
   enum state state = NODATA;
+  BPFtrace* bpftrace_ = nullptr;
 };
 
 inline bool BTF::has_data(void) const
