@@ -1,16 +1,23 @@
 #pragma once
 
-#include <string>
 #include <memory>
+#include <string>
 
 namespace bpftrace {
 
-enum class AlignState { Ok, Fail, NotAlign, NotSupp };
+enum class AlignState
+{
+  Ok,
+  Fail,
+  NotAlign,
+  NotSupp
+};
 
 class IDisasm
 {
 public:
   virtual AlignState is_aligned(uint64_t offset, uint64_t pc) = 0;
+  virtual ~IDisasm() = default;
 };
 
 class Disasm
