@@ -3,13 +3,10 @@
 #include <fstream>
 #include <memory>
 
-#include "ast.h"
+#include "ast/ast.h"
 #include "bpftrace.h"
-#include "parser.tab.hh"
 
 typedef void* yyscan_t;
-#define YY_DECL bpftrace::Parser::symbol_type yylex(bpftrace::Driver &driver, yyscan_t yyscanner)
-YY_DECL;
 
 namespace bpftrace {
 
@@ -25,15 +22,21 @@ public:
   void error(std::ostream &, const location &, const std::string &);
   void error(const location &l, const std::string &m);
   void error(const std::string &m);
-  ast::Program *root_{ nullptr };
+  ast::Program *root_ = nullptr;
+
+  void debug()
+  {
+    debug_ = true;
+  };
 
   BPFtrace &bpftrace_;
 
+  bool listing_ = false;
+
 private:
-  std::unique_ptr<Parser> parser_;
   std::ostream &out_;
-  yyscan_t scanner_;
   bool failed_ = false;
+  bool debug_ = false;
 };
 
 } // namespace bpftrace
