@@ -1,27 +1,39 @@
+#include <array>
 #include <csignal>
 #include <cstdio>
 #include <cstring>
 #include <fstream>
 #include <getopt.h>
 #include <iostream>
+#include <limits>
+#include <optional>
 #include <sys/resource.h>
 #include <sys/utsname.h>
+#include <time.h>
 #include <unistd.h>
 
+#include "aot/aot.h"
+#include "ast/bpforc/bpforc.h"
+#include "ast/pass_manager.h"
+
+#include "ast/passes/codegen_llvm.h"
+#include "ast/passes/field_analyser.h"
+#include "ast/passes/node_counter.h"
+#include "ast/passes/portability_analyser.h"
+#include "ast/passes/resource_analyser.h"
+#include "ast/passes/semantic_analyser.h"
+
 #include "bpffeature.h"
-#include "bpforc.h"
 #include "bpftrace.h"
+#include "build_info.h"
 #include "child.h"
 #include "clang_parser.h"
-#include "codegen_llvm.h"
 #include "driver.h"
-#include "field_analyser.h"
-#include "list.h"
 #include "lockdown.h"
+#include "log.h"
 #include "output.h"
-#include "printer.h"
+#include "probe_matcher.h"
 #include "procmon.h"
-#include "semantic_analyser.h"
 #include "tracepoint_format_parser.h"
 
 using namespace bpftrace;
@@ -33,6 +45,21 @@ enum class OutputBufferConfig {
   FULL,
   NONE,
 };
+
+enum class TestMode
+{
+  UNSET = 0,
+  SEMANTIC,
+  CODEGEN,
+};
+
+enum class BuildMode
+{
+  // Compile script and run immediately
+  DYNAMIC = 0,
+  // Compile script into portable executable
+  AHEAD_OF_TIME,
+};
 } // namespace
 
 void usage()
@@ -49,7 +76,6 @@ void usage()
   std::cerr << "    -o file        redirect bpftrace output to file" << std::endl;
   std::cerr << "    -d             debug info dry run" << std::endl;
   std::cerr << "    -dd            verbose debug info dry run" << std::endl;
-  std::cerr << "    -b             force BTF (BPF type format) processing" << std::endl;
   std::cerr << "    -e 'program'   execute this program" << std::endl;
   std::cerr << "    -h, --help     show this help message" << std::endl;
   std::cerr << "    -I DIR         add the directory to the include search path" << std::endl;
@@ -60,11 +86,14 @@ void usage()
   std::cerr << "    --usdt-file-activation" << std::endl;
   std::cerr << "                   activate usdt semaphores based on file path" << std::endl;
   std::cerr << "    --unsafe       allow unsafe builtin functions" << std::endl;
+  std::cerr << "    -q             keep messages quiet" << std::endl;
   std::cerr << "    -v             verbose messages" << std::endl;
   std::cerr << "    --info         Print information about kernel BPF support" << std::endl;
   std::cerr << "    -k             emit a warning when a bpf helper returns an error (except read functions)" << std::endl;
   std::cerr << "    -kk            check all bpf helper functions" << std::endl;
-  std::cerr << "    -V, --version  bpftrace version" << std::endl << std::endl;
+  std::cerr << "    -V, --version  bpftrace version" << std::endl;
+  std::cerr << "    --no-warnings  disable all warning messages" << std::endl;
+  std::cerr << std::endl;
   std::cerr << "ENVIRONMENT:" << std::endl;
   std::cerr << "    BPFTRACE_STRLEN             [default: 64] bytes on BPF stack per str()" << std::endl;
   std::cerr << "    BPFTRACE_NO_CPP_DEMANGLE    [default: 0] disable C++ symbol demangling" << std::endl;
@@ -96,16 +125,16 @@ static void enforce_infinite_rlimit() {
   rl.rlim_cur = rl.rlim_max;
   err = setrlimit(RLIMIT_MEMLOCK, &rl);
   if (err)
-    std::cerr << std::strerror(err)<<": couldn't set RLIMIT_MEMLOCK for " <<
-        "bpftrace. If your program is not loading, you can try " <<
-        "\"ulimit -l 8192\" to fix the problem" << std::endl;
+    LOG(ERROR) << std::strerror(err) << ": couldn't set RLIMIT_MEMLOCK for "
+               << "bpftrace. If your program is not loading, you can try "
+               << "\"ulimit -l 8192\" to fix the problem";
 }
 
 bool is_root()
 {
   if (geteuid() != 0)
   {
-    std::cerr << "ERROR: bpftrace currently only supports running as the root user." << std::endl;
+    LOG(ERROR) << "bpftrace currently only supports running as the root user.";
     return false;
   }
   else
@@ -122,87 +151,301 @@ static int info()
             << utsname.version << std::endl
             << "  Arch: " << utsname.machine << std::endl;
 
-  std::cerr << std::endl
-            << "Build" << std::endl
-            << "  version: " << BPFTRACE_VERSION << std::endl
-            << "  LLVM: " << LLVM_VERSION_MAJOR << std::endl
-            << "  foreach_sym: "
-#ifdef HAVE_BCC_ELF_FOREACH_SYM
-            << "yes" << std::endl
-#else
-            << "no" << std::endl
-#endif
-            << "  unsafe uprobe: "
-#ifdef HAVE_UNSAFE_UPROBE
-            << "yes" << std::endl;
-#else
-            << "no" << std::endl;
-#endif
-  std::cerr << "  bfd: "
-#ifdef HAVE_BFD_DISASM
-            << "yes" << std::endl;
-#else
-            << "no" << std::endl;
-#endif
-  std::cerr << "  bpf_attach_kfunc: "
-#ifdef HAVE_BCC_KFUNC
-            << "yes" << std::endl;
-#else
-            << "no" << std::endl;
-#endif
-  std::cerr << "  bcc_usdt_addsem: "
-#ifdef HAVE_BCC_USDT_ADDSEM
-            << "yes" << std::endl;
-#else
-            << "no" << std::endl;
-#endif
-  std::cerr << "  bcc bpf_attach_uprobe refcount: "
-#ifdef LIBBCC_ATTACH_UPROBE_SEVEN_ARGS_SIGNATURE
-            << "yes" << std::endl;
-#else
-            << "no" << std::endl;
-#endif
-  std::cerr << "  libbpf: "
-#ifdef HAVE_LIBBPF
-            << "yes" << std::endl;
-#else
-            << "no" << std::endl;
-#endif
-  std::cerr << "  libbpf btf dump: "
-#ifdef HAVE_LIBBPF_BTF_DUMP
-            << "yes" << std::endl;
-#else
-            << "no" << std::endl;
-#endif
-  std::cerr << "  libbpf btf dump type decl: "
-#ifdef HAVE_LIBBPF_BTF_DUMP_EMIT_TYPE_DECL
-            << "yes" << std::endl;
-#else
-            << "no" << std::endl;
-#endif
+  std::cerr << std::endl;
+  std::cerr << BuildInfo::report();
 
   std::cerr << std::endl;
-
   std::cerr << BPFfeature().report();
 
   return 0;
 }
 
-int main(int argc, char *argv[])
+static std::optional<struct timespec> get_boottime()
+{
+  std::optional<struct timespec> ret = std::nullopt;
+  long lowest_delta = std::numeric_limits<long>::max();
+
+  // Run the "triple vdso sandwich" 5 times, taking the result from the
+  // iteration with the lowest delta between first and last clock_gettime()
+  // calls.
+  for (int i = 0; i < 5; ++i)
+  {
+    struct timespec before, after, boottime;
+    long delta;
+
+    if (::clock_gettime(CLOCK_REALTIME, &before))
+      continue;
+
+    if (::clock_gettime(CLOCK_BOOTTIME, &boottime))
+      continue;
+
+    if (::clock_gettime(CLOCK_REALTIME, &after))
+      continue;
+
+    // There's no way 3 VDSO calls should take more than 1s. We'll
+    // also ignore the case where we cross a 1s boundary b/c that
+    // can only happen once and we're running this loop 5 times.
+    // This helps keep the math simple.
+    if (before.tv_sec != after.tv_sec)
+      continue;
+
+    delta = after.tv_nsec - before.tv_nsec;
+
+    // Time went backwards
+    if (delta < 0)
+      continue;
+
+    // Lowest delta seen so far, compute boot realtime and store it
+    if (delta < lowest_delta)
+    {
+      struct timespec boottime_realtime;
+      long nsec_avg = (before.tv_nsec + after.tv_nsec) / 2;
+      if (nsec_avg - boottime.tv_nsec < 0)
+      {
+        boottime_realtime.tv_sec = after.tv_sec - boottime.tv_sec - 1;
+        boottime_realtime.tv_nsec = nsec_avg - boottime.tv_nsec + 1e9;
+      }
+      else
+      {
+        boottime_realtime.tv_sec = after.tv_sec - boottime.tv_sec;
+        boottime_realtime.tv_nsec = nsec_avg - boottime.tv_nsec;
+      }
+
+      lowest_delta = delta;
+      ret = boottime_realtime;
+    }
+  }
+
+  if (ret && lowest_delta >= 1e5)
+    LOG(WARNING) << (lowest_delta / 1e3)
+                 << "us skew detected when calculating boot time. strftime() "
+                    "builtin may be inaccurate";
+
+  return ret;
+}
+
+[[nodiscard]] static bool parse_env(BPFtrace& bpftrace)
+{
+  if (!get_uint64_env_var("BPFTRACE_STRLEN", bpftrace.strlen_))
+    return false;
+
+  // in practice, the largest buffer I've seen fit into the BPF stack was 240
+  // bytes. I've set the bar lower, in case your program has a deeper stack than
+  // the one from my tests, in the hope that you'll get this instructive error
+  // instead of getting the BPF verifier's error.
+  if (bpftrace.strlen_ > 200)
+  {
+    // the verifier errors you would encounter when attempting larger
+    // allocations would be: >240=  <Looks like the BPF stack limit of 512 bytes
+    // is exceeded. Please move large on stack variables into BPF per-cpu array
+    // map.> ~1024= <A call to built-in function 'memset' is not supported.>
+    LOG(ERROR) << "'BPFTRACE_STRLEN' " << bpftrace.strlen_
+               << " exceeds the current maximum of 200 bytes.\n"
+               << "This limitation is because strings are currently stored on "
+                  "the 512 byte BPF stack.\n"
+               << "Long strings will be pursued in: "
+                  "https://github.com/iovisor/bpftrace/issues/305";
+    return false;
+  }
+
+  if (const char* env_p = std::getenv("BPFTRACE_NO_CPP_DEMANGLE"))
+  {
+    if (std::string(env_p) == "1")
+      bpftrace.demangle_cpp_symbols_ = false;
+    else if (std::string(env_p) == "0")
+      bpftrace.demangle_cpp_symbols_ = true;
+    else
+    {
+      LOG(ERROR) << "Env var 'BPFTRACE_NO_CPP_DEMANGLE' did not contain a "
+                    "valid value (0 or 1).";
+      return false;
+    }
+  }
+
+  if (!get_uint64_env_var("BPFTRACE_MAP_KEYS_MAX", bpftrace.mapmax_))
+    return false;
+
+  if (!get_uint64_env_var("BPFTRACE_MAX_PROBES", bpftrace.max_probes_))
+    return false;
+
+  if (!get_uint64_env_var("BPFTRACE_LOG_SIZE", bpftrace.log_size_))
+    return false;
+
+  if (!get_uint64_env_var("BPFTRACE_PERF_RB_PAGES", bpftrace.perf_rb_pages_))
+    return false;
+
+  if (!get_uint64_env_var("BPFTRACE_MAX_TYPE_RES_ITERATIONS",
+                          bpftrace.max_type_res_iterations))
+    return 1;
+
+  if (!get_uint64_env_var("BPFTRACE_MAX_TYPE_RES_ITERATIONS",
+                          bpftrace.max_type_res_iterations))
+    return false;
+
+  if (const char* env_p = std::getenv("BPFTRACE_CAT_BYTES_MAX"))
+  {
+    uint64_t proposed;
+    std::istringstream stringstream(env_p);
+    if (!(stringstream >> proposed))
+    {
+      LOG(ERROR) << "Env var 'BPFTRACE_CAT_BYTES_MAX' did not contain a valid "
+                    "uint64_t, or was zero-valued.";
+      return false;
+    }
+    bpftrace.cat_bytes_max_ = proposed;
+  }
+
+  if (const char* env_p = std::getenv("BPFTRACE_NO_USER_SYMBOLS"))
+  {
+    std::string s(env_p);
+    if (s == "1")
+      bpftrace.resolve_user_symbols_ = false;
+    else if (s == "0")
+      bpftrace.resolve_user_symbols_ = true;
+    else
+    {
+      LOG(ERROR) << "Env var 'BPFTRACE_NO_USER_SYMBOLS' did not contain a "
+                    "valid value (0 or 1).";
+      return false;
+    }
+  }
+
+  if (const char* env_p = std::getenv("BPFTRACE_CACHE_USER_SYMBOLS"))
+  {
+    std::string s(env_p);
+    if (s == "1")
+      bpftrace.cache_user_symbols_ = true;
+    else if (s == "0")
+      bpftrace.cache_user_symbols_ = false;
+    else
+    {
+      LOG(ERROR) << "Env var 'BPFTRACE_CACHE_USER_SYMBOLS' did not contain a "
+                    "valid value (0 or 1).";
+      return false;
+    }
+  }
+  else
+  {
+    // enable user symbol cache if ASLR is disabled on system or `-c` option is
+    // given
+    bpftrace.cache_user_symbols_ = !bpftrace.cmd_.empty() ||
+                                   !bpftrace.is_aslr_enabled(-1);
+  }
+
+  uint64_t node_max = std::numeric_limits<uint64_t>::max();
+  if (!get_uint64_env_var("BPFTRACE_NODE_MAX", node_max))
+    return false;
+
+  bpftrace.ast_max_nodes_ = node_max;
+  return true;
+}
+
+[[nodiscard]] std::unique_ptr<ast::Node> parse(
+    BPFtrace& bpftrace,
+    const std::string& name,
+    const std::string& program,
+    const std::vector<std::string>& include_dirs,
+    const std::vector<std::string>& include_files)
+{
+  Driver driver(bpftrace);
+  driver.source(name, program);
+  int err;
+
+  err = driver.parse();
+  if (err)
+    return nullptr;
+
+  ast::FieldAnalyser fields(driver.root_, bpftrace);
+  err = fields.analyse();
+  if (err)
+    return nullptr;
+
+  if (TracepointFormatParser::parse(driver.root_, bpftrace) == false)
+    return nullptr;
+
+  ClangParser clang;
+  std::vector<std::string> extra_flags;
+  {
+    struct utsname utsname;
+    uname(&utsname);
+    std::string ksrc, kobj;
+    auto kdirs = get_kernel_dirs(utsname, !bpftrace.feature_->has_btf());
+    ksrc = std::get<0>(kdirs);
+    kobj = std::get<1>(kdirs);
+
+    if (ksrc != "")
+      extra_flags = get_kernel_cflags(utsname.machine, ksrc, kobj);
+  }
+  extra_flags.push_back("-include");
+  extra_flags.push_back(CLANG_WORKAROUNDS_H);
+
+  for (auto dir : include_dirs)
+  {
+    extra_flags.push_back("-I");
+    extra_flags.push_back(dir);
+  }
+  for (auto file : include_files)
+  {
+    extra_flags.push_back("-include");
+    extra_flags.push_back(file);
+  }
+
+  // NOTE(mmarchini): if there are no C definitions, clang parser won't run to
+  // avoid issues in some versions. Since we're including files in the command
+  // line, we want to force parsing, so we make sure C definitions are not
+  // empty before going to clang parser stage.
+  if (!include_files.empty() && driver.root_->c_definitions.empty())
+    driver.root_->c_definitions = "#define __BPFTRACE_DUMMY__";
+
+  if (!clang.parse(driver.root_, bpftrace, extra_flags))
+    return nullptr;
+
+  err = driver.parse();
+  if (err)
+    return nullptr;
+
+  auto ast = driver.root_;
+  driver.root_ = nullptr;
+  return std::unique_ptr<ast::Node>(ast);
+}
+
+ast::PassManager CreateDynamicPM()
+{
+  ast::PassManager pm;
+  pm.AddPass(ast::CreateSemanticPass());
+  pm.AddPass(ast::CreateCounterPass());
+  pm.AddPass(ast::CreateResourcePass());
+
+  return pm;
+}
+
+ast::PassManager CreateAotPM(std::string __attribute__((unused)))
+{
+  ast::PassManager pm;
+  pm.AddPass(ast::CreateSemanticPass());
+  pm.AddPass(ast::CreatePortabilityPass());
+  pm.AddPass(ast::CreateResourcePass());
+
+  return pm;
+}
+
+int main(int argc, char* argv[])
 {
   int err;
   std::string pid_str;
   std::string cmd_str;
   bool listing = false;
   bool safe_mode = true;
-  bool force_btf = false;
   bool usdt_file_activation = false;
   int helper_check_level = 0;
-  std::string script, search, file_name, output_file, output_format;
+  TestMode test_mode = TestMode::UNSET;
+  std::string script, search, file_name, output_file, output_format, output_elf,
+      aot;
   OutputBufferConfig obc = OutputBufferConfig::UNSET;
+  BuildMode build_mode = BuildMode::DYNAMIC;
   int c;
 
-  const char* const short_options = "dbB:f:e:hlp:vc:Vo:I:k";
+  const char* const short_options = "dbB:f:e:hlp:vqc:Vo:I:k";
   option long_options[] = {
     option{ "help", no_argument, nullptr, 'h' },
     option{ "version", no_argument, nullptr, 'V' },
@@ -211,6 +454,10 @@ int main(int argc, char *argv[])
     option{ "btf", no_argument, nullptr, 'b' },
     option{ "include", required_argument, nullptr, '#' },
     option{ "info", no_argument, nullptr, 2000 },
+    option{ "emit-elf", required_argument, nullptr, 2001 },
+    option{ "no-warnings", no_argument, nullptr, 2002 },
+    option{ "test", required_argument, nullptr, 2003 },
+    option{ "aot", required_argument, nullptr, 2004 },
     option{ nullptr, 0, nullptr, 0 }, // Must be last
   };
   std::vector<std::string> include_dirs;
@@ -225,6 +472,27 @@ int main(int argc, char *argv[])
           return info();
         return 1;
         break;
+      case 2001: // --emit-elf
+        output_elf = optarg;
+        break;
+      case 2002: // --no-warnings
+        DISABLE_LOG(WARNING);
+        break;
+      case 2003: // --test
+        if (std::strcmp(optarg, "semantic") == 0)
+          test_mode = TestMode::SEMANTIC;
+        else if (std::strcmp(optarg, "codegen") == 0)
+          test_mode = TestMode::CODEGEN;
+        else
+        {
+          LOG(ERROR) << "USAGE: --test must be either 'semantic' or 'codegen'.";
+          return 1;
+        }
+        break;
+      case 2004: // --aot
+        aot = optarg;
+        build_mode = BuildMode::AHEAD_OF_TIME;
+        break;
       case 'o':
         output_file = optarg;
         break;
@@ -235,6 +503,9 @@ int main(int argc, char *argv[])
           return 1;
         }
         break;
+      case 'q':
+        bt_quiet = true;
+        break;
       case 'v':
         bt_verbose = true;
         break;
@@ -246,7 +517,7 @@ int main(int argc, char *argv[])
         } else if (std::strcmp(optarg, "none") == 0) {
           obc = OutputBufferConfig::NONE;
         } else {
-          std::cerr << "USAGE: -B must be either 'line', 'full', or 'none'." << std::endl;
+          LOG(ERROR) << "USAGE: -B must be either 'line', 'full', or 'none'.";
           return 1;
         }
         break;
@@ -278,7 +549,6 @@ int main(int argc, char *argv[])
         safe_mode = false;
         break;
       case 'b':
-        force_btf = true;
         break;
       case 'h':
         usage();
@@ -308,13 +578,13 @@ int main(int argc, char *argv[])
   if (bt_verbose && (bt_debug != DebugLevel::kNone))
   {
     // TODO: allow both
-    std::cerr << "USAGE: Use either -v or -d." << std::endl;
+    LOG(ERROR) << "USAGE: Use either -v or -d.";
     return 1;
   }
 
   if (!cmd_str.empty() && !pid_str.empty())
   {
-    std::cerr << "USAGE: Cannot use both -c and -p." << std::endl;
+    LOG(ERROR) << "USAGE: Cannot use both -c and -p.";
     usage();
     return 1;
   }
@@ -324,8 +594,8 @@ int main(int argc, char *argv[])
   if (!output_file.empty()) {
     outputstream.open(output_file);
     if (outputstream.fail()) {
-      std::cerr << "Failed to open output file: \"" << output_file;
-      std::cerr << "\": " << strerror(errno) <<  std::endl;
+      LOG(ERROR) << "Failed to open output file: \"" << output_file
+                 << "\": " << strerror(errno);
       return 1;
     }
     os = &outputstream;
@@ -339,8 +609,8 @@ int main(int argc, char *argv[])
     output = std::make_unique<JsonOutput>(*os);
   }
   else {
-    std::cerr << "Invalid output format \"" << output_format << "\"" << std::endl;
-    std::cerr << "Valid formats: 'text', 'json'" << std::endl;
+    LOG(ERROR) << "Invalid output format \"" << output_format << "\"\n"
+               << "Valid formats: 'text', 'json'";
     return 1;
   }
 
@@ -355,18 +625,27 @@ int main(int argc, char *argv[])
     case OutputBufferConfig::NONE:
       std::setvbuf(stdout, NULL, _IONBF, BUFSIZ);
       break;
-    default:
-      // Should never get here
-      std::abort();
   }
 
   BPFtrace bpftrace(std::move(output));
-  Driver driver(bpftrace);
+
+  if (!cmd_str.empty())
+    bpftrace.cmd_ = cmd_str;
+
+  if (!parse_env(bpftrace))
+    return 1;
+
+  // Difficult to serialize flex generated types
+  if (helper_check_level && build_mode == BuildMode::AHEAD_OF_TIME)
+  {
+    LOG(ERROR) << "Cannot use -k[k] with --aot";
+    return 1;
+  }
 
   bpftrace.usdt_file_activation_ = usdt_file_activation;
   bpftrace.safe_mode_ = safe_mode;
-  bpftrace.force_btf_ = force_btf;
   bpftrace.helper_check_level_ = helper_check_level;
+  bpftrace.boottime_ = get_boottime();
 
   if (!pid_str.empty())
   {
@@ -376,36 +655,79 @@ int main(int argc, char *argv[])
     }
     catch (const std::exception& e)
     {
-      std::cerr << "ERROR: " << e.what() << std::endl;
+      LOG(ERROR) << e.what();
       return 1;
     }
   }
 
+  if (!cmd_str.empty())
+  {
+    bpftrace.cmd_ = cmd_str;
+    try
+    {
+      bpftrace.child_ = std::make_unique<ChildProc>(cmd_str);
+    }
+    catch (const std::runtime_error& e)
+    {
+      LOG(ERROR) << "Failed to fork child: " << e.what();
+      return -1;
+    }
+  }
+
   // Listing probes
   if (listing)
   {
     if (!is_root())
       return 1;
 
-    if (optind == argc-1)
-      list_probes(bpftrace, argv[optind]);
-    else if (optind == argc)
-      list_probes(bpftrace, "");
+    if (optind == argc || std::string(argv[optind]) == "*")
+      script = "*:*";
+    else if (optind == argc - 1)
+      script = argv[optind];
     else
+    {
       usage();
+      return 1;
+    }
 
+    if (script.find(':') == std::string::npos &&
+        (script.find("struct") == 0 || script.find("union") == 0 ||
+         script.find("enum") == 0))
+    {
+      // Print structure definitions
+      bpftrace.probe_matcher_->list_structs(script);
+      return 0;
+    }
+
+    Driver driver(bpftrace);
+    driver.listing_ = true;
+    driver.source("stdin", script);
+
+    int err = driver.parse();
+    if (err)
+      return err;
+
+    ast::SemanticAnalyser semantics(driver.root_, bpftrace, false, true);
+    err = semantics.analyse();
+    if (err)
+      return err;
+
+    bpftrace.probe_matcher_->list_probes(driver.root_);
     return 0;
   }
 
+  std::string filename;
+  std::string program;
+
   if (script.empty())
   {
     // Script file
     if (argv[optind] == nullptr)
     {
-      std::cerr << "USAGE: filename or -e 'program' required." << std::endl;
+      LOG(ERROR) << "USAGE: filename or -e 'program' required.";
       return 1;
     }
-    std::string filename(argv[optind]);
+    filename = argv[optind];
     std::stringstream buf;
 
     if (filename == "-")
@@ -419,20 +741,22 @@ int main(int argc, char *argv[])
         buf << line << std::endl;
       }
 
-      driver.source("stdin", buf.str());
+      filename = "stdin";
+      program = buf.str();
     }
     else
     {
       std::ifstream file(filename);
       if (file.fail())
       {
-        std::cerr << "Error opening file '" << filename << "': ";
-        std::cerr << std::strerror(errno) << std::endl;
+        LOG(ERROR) << "failed to open file '" << filename
+                   << "': " << std::strerror(errno);
         return -1;
       }
 
+      program = buf.str();
       buf << file.rdbuf();
-      driver.source(filename, buf.str());
+      program = buf.str();
     }
 
     optind++;
@@ -440,7 +764,8 @@ int main(int argc, char *argv[])
   else
   {
     // Script is provided as a command line argument
-    driver.source("stdin", script);
+    filename = "stdin";
+    program = script;
   }
 
   // Load positional parameters before driver runs so positional
@@ -451,10 +776,6 @@ int main(int argc, char *argv[])
     optind++;
   }
 
-  err = driver.parse();
-  if (err)
-    return err;
-
   if (!is_root())
     return 1;
 
@@ -465,184 +786,34 @@ int main(int argc, char *argv[])
     return 1;
   }
 
-  ast::FieldAnalyser fields(driver.root_, bpftrace);
-  err = fields.analyse();
-  if (err)
-    return err;
-
   // FIXME (mmarchini): maybe we don't want to always enforce an infinite
   // rlimit?
   enforce_infinite_rlimit();
 
-  // defaults
-  bpftrace.join_argnum_ = 16;
-  bpftrace.join_argsize_ = 1024;
-
-  if (!get_uint64_env_var("BPFTRACE_STRLEN", bpftrace.strlen_))
+  auto ast_root = parse(
+      bpftrace, filename, program, include_dirs, include_files);
+  if (!ast_root)
     return 1;
 
-  // in practice, the largest buffer I've seen fit into the BPF stack was 240 bytes.
-  // I've set the bar lower, in case your program has a deeper stack than the one from my tests,
-  // in the hope that you'll get this instructive error instead of getting the BPF verifier's error.
-  if (bpftrace.strlen_ > 200) {
-    // the verifier errors you would encounter when attempting larger allocations would be:
-    // >240=  <Looks like the BPF stack limit of 512 bytes is exceeded. Please move large on stack variables into BPF per-cpu array map.>
-    // ~1024= <A call to built-in function 'memset' is not supported.>
-    std::cerr << "'BPFTRACE_STRLEN' " << bpftrace.strlen_ << " exceeds the current maximum of 200 bytes." << std::endl
-    << "This limitation is because strings are currently stored on the 512 byte BPF stack." << std::endl
-    << "Long strings will be pursued in: https://github.com/iovisor/bpftrace/issues/305" << std::endl;
-    return 1;
-  }
-
-  if (const char* env_p = std::getenv("BPFTRACE_NO_CPP_DEMANGLE"))
+  ast::PassContext ctx(bpftrace);
+  ast::PassManager pm;
+  switch (build_mode)
   {
-    if (std::string(env_p) == "1")
-      bpftrace.demangle_cpp_symbols_ = false;
-    else if (std::string(env_p) == "0")
-      bpftrace.demangle_cpp_symbols_ = true;
-    else
-    {
-      std::cerr << "Env var 'BPFTRACE_NO_CPP_DEMANGLE' did not contain a valid value (0 or 1)." << std::endl;
-      return 1;
-    }
+    case BuildMode::DYNAMIC:
+      pm = CreateDynamicPM();
+      break;
+    case BuildMode::AHEAD_OF_TIME:
+      pm = CreateAotPM(aot);
+      break;
   }
 
-  if (!get_uint64_env_var("BPFTRACE_MAP_KEYS_MAX", bpftrace.mapmax_))
+  auto pmresult = pm.Run(std::move(ast_root), ctx);
+  if (!pmresult.Ok())
     return 1;
 
-  if (!get_uint64_env_var("BPFTRACE_MAX_PROBES", bpftrace.max_probes_))
-    return 1;
+  ast_root = std::unique_ptr<ast::Node>(pmresult.Root());
 
-  if (!get_uint64_env_var("BPFTRACE_LOG_SIZE", bpftrace.log_size_))
-    return 1;
-
-  if (!get_uint64_env_var("BPFTRACE_PERF_RB_PAGES", bpftrace.perf_rb_pages_))
-    return 1;
-
-  if (const char* env_p = std::getenv("BPFTRACE_CAT_BYTES_MAX"))
-  {
-    uint64_t proposed;
-    std::istringstream stringstream(env_p);
-    if (!(stringstream >> proposed)) {
-      std::cerr << "Env var 'BPFTRACE_CAT_BYTES_MAX' did not contain a valid uint64_t, or was zero-valued." << std::endl;
-      return 1;
-    }
-    bpftrace.cat_bytes_max_ = proposed;
-  }
-
-  if (const char* env_p = std::getenv("BPFTRACE_NO_USER_SYMBOLS"))
-  {
-    std::string s(env_p);
-    if (s == "1")
-      bpftrace.resolve_user_symbols_ = false;
-    else if (s == "0")
-      bpftrace.resolve_user_symbols_ = true;
-    else
-    {
-      std::cerr << "Env var 'BPFTRACE_NO_USER_SYMBOLS' did not contain a valid value (0 or 1)." << std::endl;
-      return 1;
-    }
-  }
-
-  if (const char* env_p = std::getenv("BPFTRACE_CACHE_USER_SYMBOLS"))
-  {
-    std::string s(env_p);
-    if (s == "1")
-      bpftrace.cache_user_symbols_ = true;
-    else if (s == "0")
-      bpftrace.cache_user_symbols_ = false;
-    else
-    {
-      std::cerr << "Env var 'BPFTRACE_CACHE_USER_SYMBOLS' did not contain a "
-                   "valid value (0 or 1)."
-                << std::endl;
-      return 1;
-    }
-  }
-  else
-  {
-    // enable user symbol cache if ASLR is disabled on system or `-c` option is
-    // given
-    bpftrace.cache_user_symbols_ = !cmd_str.empty() ||
-                                   !bpftrace.is_aslr_enabled(-1);
-  }
-
-  if (!cmd_str.empty())
-    bpftrace.cmd_ = cmd_str;
-
-  if (TracepointFormatParser::parse(driver.root_, bpftrace) == false)
-    return 1;
-
-  if (bt_debug != DebugLevel::kNone)
-  {
-    std::cout << "\nAST\n";
-    std::cout << "-------------------\n";
-    ast::Printer p(std::cout);
-    driver.root_->accept(p);
-    std::cout << std::endl;
-  }
-
-  ClangParser clang;
-  std::vector<std::string> extra_flags;
-  {
-    struct utsname utsname;
-    uname(&utsname);
-    std::string ksrc, kobj;
-    auto kdirs = get_kernel_dirs(utsname);
-    ksrc = std::get<0>(kdirs);
-    kobj = std::get<1>(kdirs);
-
-    if (ksrc != "")
-      extra_flags = get_kernel_cflags(utsname.machine, ksrc, kobj);
-  }
-  extra_flags.push_back("-include");
-  extra_flags.push_back(CLANG_WORKAROUNDS_H);
-
-  for (auto dir : include_dirs)
-  {
-    extra_flags.push_back("-I");
-    extra_flags.push_back(dir);
-  }
-  for (auto file : include_files)
-  {
-    extra_flags.push_back("-include");
-    extra_flags.push_back(file);
-  }
-
-  // NOTE(mmarchini): if there are no C definitions, clang parser won't run to
-  // avoid issues in some versions. Since we're including files in the command
-  // line, we want to force parsing, so we make sure C definitions are not
-  // empty before going to clang parser stage.
-  if (!include_files.empty() && driver.root_->c_definitions.empty())
-    driver.root_->c_definitions = "#define __BPFTRACE_DUMMY__";
-
-  if (!clang.parse(driver.root_, bpftrace, extra_flags))
-    return 1;
-
-  err = driver.parse();
-  if (err)
-    return err;
-
-  ast::SemanticAnalyser semantics(
-      driver.root_, bpftrace, bpftrace.feature_, !cmd_str.empty());
-  err = semantics.analyse();
-  if (err)
-    return err;
-
-  if (bt_debug != DebugLevel::kNone)
-  {
-    std::cout << "\nAST after semantic analysis\n";
-    std::cout << "-------------------\n";
-    ast::Printer p(std::cout, true);
-    driver.root_->accept(p);
-    std::cout << std::endl;
-  }
-
-  err = semantics.create_maps(bt_debug != DebugLevel::kNone);
-  if (err)
-    return err;
-
-  if (!cmd_str.empty())
+  if (!bpftrace.cmd_.empty())
   {
     try
     {
@@ -650,51 +821,75 @@ int main(int argc, char *argv[])
     }
     catch (const std::runtime_error& e)
     {
-      std::cerr << "Failed to fork child: " << e.what() << std::endl;
+      LOG(ERROR) << "Failed to fork child: " << e.what();
       return -1;
     }
   }
 
-  ast::CodegenLLVM llvm(driver.root_, bpftrace);
-  std::unique_ptr<BpfOrc> bpforc;
+  ast::CodegenLLVM llvm(&*ast_root, bpftrace);
+  BpfBytecode bytecode;
   try
   {
-    bpforc = llvm.compile(bt_debug);
+    std::unique_ptr<BpfOrc> bpforc;
+
+    llvm.generate_ir();
+    if (bt_debug == DebugLevel::kFullDebug)
+    {
+      std::cout << "Before optimization\n";
+      std::cout << "-------------------\n\n";
+      llvm.DumpIR();
+    }
+
+    llvm.optimize();
+    if (bt_debug != DebugLevel::kNone)
+    {
+      if (bt_debug == DebugLevel::kFullDebug)
+      {
+        std::cout << "\nAfter optimization\n";
+        std::cout << "------------------\n\n";
+      }
+      llvm.DumpIR();
+    }
+    if (!output_elf.empty())
+    {
+      llvm.emit_elf(output_elf);
+      return 0;
+    }
+    bpforc = llvm.emit();
+    if (bt_debug == DebugLevel::kFullDebug)
+    {
+      std::cout << "\nLLVM JITDLib state\n";
+      std::cout << "------------------\n\n";
+      raw_os_ostream os(std::cout);
+      bpforc->dump(os);
+    }
+
+    bytecode = bpforc->getBytecode();
+  }
+  catch (const std::system_error& ex)
+  {
+    LOG(ERROR) << "failed to write elf: " << ex.what();
+    return 1;
   }
   catch (const std::exception& ex)
   {
-    std::cerr << "Failed to compile: " << ex.what() << std::endl;
+    LOG(ERROR) << "Failed to compile: " << ex.what();
     return 1;
   }
 
-  if (bt_debug != DebugLevel::kNone)
+  if (bt_debug != DebugLevel::kNone || test_mode == TestMode::CODEGEN)
     return 0;
 
+  if (build_mode == BuildMode::AHEAD_OF_TIME)
+    return aot::generate(bpftrace.resources, bytecode, aot);
+
   // Signal handler that lets us know an exit signal was received.
   struct sigaction act = {};
   act.sa_handler = [](int) { BPFtrace::exitsig_recv = true; };
   sigaction(SIGINT, &act, NULL);
   sigaction(SIGTERM, &act, NULL);
 
-  uint64_t num_probes = bpftrace.num_probes();
-  if (num_probes == 0)
-  {
-    std::cout << "No probes to attach" << std::endl;
-    return 1;
-  }
-  else if (num_probes > bpftrace.max_probes_)
-  {
-    std::cerr << "Can't attach to " << num_probes << " probes because it "
-      << "exceeds the current limit of " << bpftrace.max_probes_ << " probes."
-      << std::endl << "You can increase the limit through the BPFTRACE_MAX_PROBES "
-      << "environment variable, but BE CAREFUL since a high number of probes "
-      << "attached can cause your system to crash." << std::endl;
-    return 1;
-  }
-  else
-    bpftrace.out_->attached_probes(num_probes);
-
-  err = bpftrace.run(move(bpforc));
+  err = bpftrace.run(bytecode);
   if (err)
     return err;
 
