@@ -1,6 +1,8 @@
 #include <cstring>
 
+#include "async_event_types.h"
 #include "bpftrace.h"
+#include "log.h"
 #include "mapkey.h"
 #include "utils.h"
 
@@ -15,7 +17,7 @@ size_t MapKey::size() const
 {
   size_t size = 0;
   for (auto &arg : args_)
-    size += arg.size;
+    size += arg.GetSize();
   return size;
 }
 
@@ -41,7 +43,7 @@ std::vector<std::string> MapKey::argument_value_list(BPFtrace &bpftrace,
   for (const SizedType &arg : args_)
   {
     list.push_back(argument_value(bpftrace, arg, &data[offset]));
-    offset += arg.size;
+    offset += arg.GetSize();
   }
   return list;
 }
@@ -63,7 +65,7 @@ std::string MapKey::argument_value(BPFtrace &bpftrace,
   switch (arg.type)
   {
     case Type::integer:
-      switch (arg.size)
+      switch (arg.GetSize())
       {
         case 1:
           return std::to_string(read_data<int8_t>(data));
@@ -83,6 +85,11 @@ std::string MapKey::argument_value(BPFtrace &bpftrace,
     case Type::ustack:
       return bpftrace.get_stack(
           read_data<uint64_t>(data), true, arg.stack_type, 4);
+    case Type::timestamp:
+    {
+      auto p = static_cast<const AsyncEvent::Strftime *>(data);
+      return bpftrace.resolve_timestamp(p->strftime_id, p->nsecs_since_boot);
+    }
     case Type::ksym:
       return bpftrace.resolve_ksym(read_data<uint64_t>(data));
     case Type::usym:
@@ -94,26 +101,52 @@ std::string MapKey::argument_value(BPFtrace &bpftrace,
     case Type::username:
       return bpftrace.resolve_uid(read_data<uint64_t>(data));
     case Type::probe:
-      return bpftrace.probe_ids_[read_data<uint64_t>(data)];
+      return bpftrace.resources.probe_ids[read_data<uint64_t>(data)];
     case Type::string:
     {
       auto p = static_cast<const char *>(data);
-      return std::string(p, strnlen(p, arg.size));
+      return std::string(p, strnlen(p, arg.GetSize()));
     }
     case Type::buffer:
     {
       auto p = static_cast<const char *>(data) + 1;
-      return hex_format_buffer(p, arg.size - 1);
+      return hex_format_buffer(p, arg.GetSize() - 1);
     }
-    case Type::cast:
-      if (arg.is_pointer) {
-        // use case: show me these pointer values
-        ptr << "0x" << std::hex << read_data<int64_t>(data);
-        return ptr.str();
+    case Type::pointer:
+    {
+      // use case: show me these pointer values
+      ptr << "0x" << std::hex << read_data<int64_t>(data);
+      return ptr.str();
+    }
+    case Type::array:
+    {
+      std::vector<std::string> elems;
+      for (size_t i = 0; i < arg.GetNumElements(); i++)
+        elems.push_back(argument_value(bpftrace,
+                                       *arg.GetElementTy(),
+                                       (const uint8_t *)data +
+                                           i * arg.GetElementTy()->GetSize()));
+      return "[" + str_join(elems, ",") + "]";
+    }
+    case Type::record:
+    {
+      std::vector<std::string> elems;
+      for (auto &field : arg.GetFields())
+      {
+        elems.push_back("." + field.name + "=" +
+                        argument_value(bpftrace,
+                                       field.type,
+                                       (const uint8_t *)data + field.offset));
       }
-      // fall through
+      return "{" + str_join(elems, ",") + "}";
+    }
+    case Type::mac_address:
+    {
+      auto p = static_cast<const uint8_t *>(data);
+      return bpftrace.resolve_mac_address(p);
+    }
     default:
-      std::cerr << "invalid mapkey argument type" << std::endl;
+      LOG(ERROR) << "invalid mapkey argument type";
   }
   abort();
 }
