@@ -1,13 +1,6 @@
 #include "gmock/gmock.h"
 #include "gtest/gtest.h"
 
-#include "bpforc.h"
-#include "bpftrace.h"
-#include "clang_parser.h"
-#include "codegen_llvm.h"
-#include "driver.h"
-#include "semantic_analyser.h"
-
 #include "common.h"
 
 namespace bpftrace {
@@ -22,10 +15,15 @@ TEST(codegen, regression_957)
   Driver driver(*bpftrace);
 
   ASSERT_EQ(driver.parse_str("t:sched:sched_one* { cat(\"%s\", probe); }"), 0);
-  MockBPFfeature feature;
-  ast::SemanticAnalyser semantics(driver.root_, *bpftrace, feature);
+  bpftrace->feature_ = std::make_unique<MockBPFfeature>(true);
+  ast::SemanticAnalyser semantics(driver.root_, *bpftrace);
   ASSERT_EQ(semantics.analyse(), 0);
-  ASSERT_EQ(semantics.create_maps(true), 0);
+
+  ast::ResourceAnalyser resource_analyser(driver.root_);
+  auto resources = resource_analyser.analyse();
+  ASSERT_EQ(resources.create_maps(*bpftrace, true), 0);
+  bpftrace->resources = resources;
+
   ast::CodegenLLVM codegen(driver.root_, *bpftrace);
   codegen.compile();
 }
