@@ -87,6 +87,17 @@ static void symlink_test_binary(const std::string& destination)
   }
 }
 
+static std::string get_working_path()
+{
+  char cwd_path[PATH_MAX];
+  if (::getcwd(cwd_path, PATH_MAX) == nullptr)
+  {
+    throw std::runtime_error(
+        "getting current working directory for tests failed");
+  }
+  return std::string(cwd_path);
+}
+
 TEST(utils, resolve_binary_path)
 {
   std::string path = "/tmp/bpftrace-test-utils-XXXXXX";
@@ -118,6 +129,35 @@ TEST(utils, resolve_binary_path)
   exec_system(("rm -rf " + path).c_str());
 }
 
+TEST(utils, abs_path)
+{
+  std::string path = "/tmp/bpftrace-test-utils-XXXXXX";
+  std::string rel_file = "bpftrace-test-utils-abs-path";
+  if (::mkdtemp(&path[0]) == nullptr)
+  {
+    throw std::runtime_error("creating temporary path for tests failed");
+  }
+
+  int fd;
+  fd = open((path + "/somefile").c_str(), O_CREAT, S_IRUSR);
+  close(fd);
+  fd = open(rel_file.c_str(), O_CREAT, S_IRUSR);
+  close(fd);
+
+  // Translates absolute path with '../..'
+  EXPECT_EQ(abs_path(path + "/../.." + path + "/somefile"), path + "/somefile");
+  // Translates relative path with './'
+  EXPECT_EQ(abs_path("./" + rel_file), get_working_path() + "/" + rel_file);
+
+  // /proc/<pid>/root path returned as is (and doesn't throw)
+  EXPECT_NO_THROW(
+      abs_path(std::string("/proc/1/root/usr/local/bin/usdt_test.so")));
+  EXPECT_EQ(abs_path(std::string("/proc/1/root/usr/local/bin/usdt_test.so")),
+            std::string("/proc/1/root/usr/local/bin/usdt_test.so"));
+
+  remove(rel_file.c_str());
+}
+
 } // namespace utils
 } // namespace test
 } // namespace bpftrace
